#!/usr/bin/env python3

##
# @file example.py
# @brief simple example using yac python bindings
#
# @copyright Copyright  (C)  2022 DKRZ, MPI-M
#
# @author Nils-Arne Dreier <dreier@dkrz.de>
#

#
# Keywords:
# Maintainer: Nils-Arne Dreier <dreier@dkrz.de>
# URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
#
# This file is part of YAC.
#
# Redistribution and use in source and binary forms, with or without
# modification, are  permitted provided that the following conditions are
# met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# Neither the name of the DKRZ GmbH nor the names of its contributors
# may be used to endorse or promote products derived from this software
# without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from yac import *
import sys
import numpy as np
try:
    import matplotlib.pyplot as plt
except:
    pass

version = version()
rank = int(sys.argv[1])
print(f"Moin im rank {rank}! I'm using YAC version {version}")

yac = YAC("coupling.xml", "coupling.xsd")

print(f"timespan: {yac.start_datetime} to {yac.end_datetime}")

comp = yac.def_comp(f"C{rank}")

x = np.linspace(-np.pi,np.pi,100)
y = np.linspace(-0.5*np.pi,0.5*np.pi,50)

grid = Reg2dGrid(f"grid{rank}", x, y)
print(f"The grid has {grid.nbr_corners} corners, {grid.nbr_edges} edges and {grid.nbr_cells} cells")

points = grid.def_points(Location.CORNER, x, y)
print(f"The points have size {points.size}")

grid.set_global_index(range(grid.nbr_corners), Location.CORNER)

field = Field.create("example_field", comp, points)

yac.search()

print(f"The role of field is {field.role}")
print(f"The timesteps are coupling: {field.coupling_timestep} - model: {field.model_timestep}")

if rank == 0:
    xv, yv = np.meshgrid(x, y, indexing="xy")
    buf = np.exp(-(xv**2+yv**2))
    print(f"Putting {field.name}...")
    field.put(buf.reshape((1,-1)))
else:
    print(f"Getting {field.name}...")
    buf, info = field.get()
    try:
        plt.imshow(buf.reshape((len(y),len(x))))
        plt.savefig("example.png")
    except:
        pass
