##
# @file yac.pyx
# @brief cython wrapper for the yac c-API
#
# Contains lightweight cython wrappers for the yac c-API.
#
# @copyright Copyright  (C)  2022 DKRZ, MPI-M
#
# @author Nils-Arne Dreier <dreier@dkrz.de>
#

#
# Keywords:
# Maintainer: Nils-Arne Dreier <dreier@dkrz.de>
# URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
#
# This file is part of YAC.
#
# Redistribution and use in source and binary forms, with or without
# modification, are  permitted provided that the following conditions are
# met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# Neither the name of the DKRZ GmbH nor the names of its contributors
# may be used to endorse or promote products derived from this software
# without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from enum import Enum
cimport numpy as np
import numpy as np

cdef import from "<mpi.h>" nogil:
  ctypedef struct _mpi_comm_t
  ctypedef _mpi_comm_t* MPI_Comm
  int MPI_Comm_c2f(MPI_Comm)

cdef extern from "yac_interface.h":
  cdef const int _CELL "YAC_LOCATION_CELL"
  cdef const int _CORNER "YAC_LOCATION_CORNER"
  cdef const int _EDGE "YAC_LOCATION_EDGE"
  void yac_cinit_instance ( const char * xml_filename,
                            const char * schema_filename,
                            int * yac_instance_id )
  void yac_cfinalize_instance (int yac_instance_id)
  void yac_cdef_comp_instance ( int yac_instance_id,
                                const char * comp_name,
                                int * comp_id )
  void yac_cget_localcomm ( MPI_Comm* local_communicator, int comp_id )
  void yac_cdef_datetime_instance ( int yac_instance_id,
                                       const char * start_datetime,
                                       const char * end_datetime )
  void yac_cget_pair_rootcomm ( MPI_Comm *local_communicator,
                                const char * comp_name1,
                                const char * comp_name2 )
  void yac_csearch_instance ( int yac_instance_id, int * ierror )
  char * yac_cget_start_datetime_instance ( int yac_instance_id )
  char * yac_cget_end_datetime_instance ( int yac_instance_id )
  char * yac_cget_version ()
  void yac_cdef_grid_reg2d ( const char * grid_name,
                             int nbr_vertices[2],
                             int cyclic[2],
                             double *x_vertices,
                             double *y_vertices,
                             int *grid_id)
  void yac_cdef_points_reg2d ( const int grid_id,
                               int nbr_points[2],
                               const int location,
                               const double *x_points,
                               const double *y_points,
                               int *point_id )
  void yac_cdef_grid_unstruct ( const char * grid_name,
                                int nbr_vertices,
                                int nbr_cells,
                                int *num_vertices_per_cell,
                                double *x_vertices,
                                double *y_vertices,
                                int *cell_to_vertex,
                                int *grid_id)
  void yac_cdef_points_unstruct ( const int grid_id,
                                  const int nbr_points,
                                  const int location,
                                  const double *x_points,
                                  const double *y_points,
                                  int *point_id )
  void yac_cset_global_index ( const int * global_index,
                               int location,
                               int grid_id)
  void yac_cdef_field ( const char * field_name,
                        const int component_id,
                        const int * point_ids,
                        const int num_pointsets,
                        int * field_id )
  void yac_cget_ ( const int field_id,
                   const int collection_size,
                   double *recv_field,
                   int    *info,
                   int    *ierror )
  void yac_cput_ ( const int field_id,
                   const int collection_size,
                   const double * send_field,
                   int *info,
                   int *ierror )
  void yac_cget_field_name_from_id ( int field_id, char *field_name )
  void yac_cget_role_of_field_by_id ( int field_id, int *role )
  void yac_cget_coupling_timestep_id ( const int field_id, char *timestep_string )
  void yac_cget_model_timestep_id ( int field_id, char *timestep_string )
  size_t yac_get_grid_size ( int location, int grid_id )
  size_t yac_get_points_size ( int points_id  )
  void yac_cget_collection_size_id ( const int field_id,
                                   int * collection_size )
class Location(Enum):
    """
    Location for points

    Refers to @ref YAC_LOCATION_CELL, @ref YAC_LOCATION_CORNER and @ref YAC_LOCATION_EDGE
    """
    CELL = _CELL
    CORNER = _CORNER
    EDGE = _EDGE

class Role(Enum):
    """
    Role for fields in a coupling
    """
    NOTHING = 0
    SOURCE = 1
    TARGET = 2

class YAC:
    """
    Initializies a YAC instance and provides further functionality

    The destructor finalizes the YAC instance by calling yac_cfinalize_instance
    """
    def __init__(self, xml_filename : str = "coupling.xml", schema_filename : str = "coupling.xsd"):
        """
        see @ref yac_cinit_instance
        """
        cdef int instance_id
        yac_cinit_instance(xml_filename.encode(), schema_filename.encode(), &instance_id)
        self.instance_id = instance_id

    def def_comp(self, comp_name : str):
        """
        see @ref yac_cdef_comp_instance
        """
        cdef int comp_id
        yac_cdef_comp_instance(self.instance_id, comp_name.encode(), &comp_id)
        return Component(comp_id)

    def def_datetime(self, start_datetime, end_datetime):
        """
        see @ref yac_cdef_datetime_instance

        The parameters can be given either as a string in iso8601
        format or as datetime objects
        """
        import datetime
        if(type(start_datetime) is datetime.datetime):
            start_datetime = start_datetime.isoformat()
        if(type(end_datetime) is datetime.datetime):
            end_datetime = end_datetime.isoformat()
        yac_cdef_datetime_instance ( self.instance_id,
                                     start_datetime.encode(),
                                     end_datetime.encode() )

    @property
    def start_datetime(self):
        """
        @see yac_cget_start_datetime_instance (`datetime.datetime`, read-only).
        """
        start = yac_cget_start_datetime_instance(self.instance_id)
        return bytes.decode(start)

    @property
    def end_datetime(self):
        """
        @see yac_cget_end_datetime_instance (`datetime.datetime`, read-only).
        """
        end = yac_cget_end_datetime_instance(self.instance_id)
        return bytes.decode(end)

    def search(self):
        """
        @see yac_csearch_instance (`datetime.datetime`, read-only).
        """
        cdef int ierror
        yac_csearch_instance(self.instance_id, &ierror)
        if ierror != 0:
            raise RuntimeError("yac_csearch returned error number " + ierror)

    def __del__(self):
        """
        @see yac_cfinalize_instance
        """
        yac_cfinalize_instance(self.instance_id)

class Component:
    """
    Stores the component_id and provides further functionality
    """
    def __init__(self, comp_id):
        self.comp_id = comp_id

    @property
    def localcomm(self):
        """
        see @ref yac_cget_localcomm (`MPI.Comm`, read-only)
        """
        from mpi4py import MPI
        cdef MPI_Comm comm
        yac_cget_localcomm(&comm, self.comp_id)
        # convert to mpi4py communicator
        return MPI.Comm.f2py(MPI_Comm_c2f(comm))

def get_pair_rootcomm(comp_name1 : str, comp_name2 : str):
    """
    see @ref yac_cget_pair_rootcomm
    """
    from mpi4py import MPI
    cdef MPI_Comm comm
    yac_cget_pair_rootcomm(&comm, comp_name1.encode(), comp_name2.encode())
    # convert to mpi4py communicator
    return MPI.Comm.f2py(MPI_Comm_c2f(comm))

class Grid:
    """
    Stores the grid_id and provides further functionality

    Base class for Reg2dGrid and UnstructuredGrid
    """
    def __init__(self, grid_id):
        self.grid_id = grid_id

    def set_global_index(self, global_index, location : Location):
        """
        see @ref yac_cset_global_index
        """
        if(type(global_index) is not np.array or global_index.dtype is not np.intc):
            global_index = np.array(global_index, dtype=np.intc)
        cdef int[:] global_index_view = global_index
        yac_cset_global_index(&global_index_view[0], location.value, self.grid_id)

    @property
    def nbr_cells(self):
        """
        see @ref yac_get_grid_size (`int`, read-only)
        """
        return yac_get_grid_size ( Location.CELL.value, self.grid_id )

    @property
    def nbr_corners(self):
        """
        see @ref yac_get_grid_size (`int`, read-only)
        """
        return yac_get_grid_size ( Location.CORNER.value, self.grid_id )

    @property
    def nbr_edges(self):
        """
        see @ref yac_get_grid_size (`int`, read-only)
        """
        return yac_get_grid_size ( Location.EDGE.value, self.grid_id )

class Points:
    """
    Stores the points_id and provides further functionality
    """
    def __init__(self, points_id):
        self.points_id = points_id

    @property
    def size(self):
        """
        see @ref yac_get_points_size (`int`, read-only)
        """
        return yac_get_points_size ( self.points_id )

class Reg2dGrid(Grid):
    """
    A stuctured 2d Grid
    """
    def __init__(self, grid_name : str, x_vertices, y_vertices,
                 cyclic = [False, False]):
        """
        see @ref yac_cdef_grid_reg2d
        """
        cdef int grid_id
        cdef double[::1] x = np.ascontiguousarray(x_vertices, dtype=np.double)
        cdef double[::1] y = np.ascontiguousarray(y_vertices, dtype=np.double)
        cdef int[2] cyclic_view = cyclic
        yac_cdef_grid_reg2d(grid_name.encode(), [len(x),len(y)],
                            cyclic_view, &x[0], &y[0], &grid_id)
        super().__init__(grid_id)

    def def_points(self, location : Location,
                   x_vertices, y_vertices):
        """
        see @ref yac_cdef_points_reg2d
        """
        cdef int points_id
        cdef double[::1] x = np.ascontiguousarray(x_vertices, dtype=np.double)
        cdef double[::1] y = np.ascontiguousarray(y_vertices, dtype=np.double)
        yac_cdef_points_reg2d(self.grid_id, [len(x), len(y)],
                              location.value, &x[0], &y[0], &points_id)
        return Points(points_id)

class UnstructuredGrid(Grid):
    """
    An unstuctured 2d Grid
    """
    def __init__(self, grid_name : str, num_vertices_per_cell,
                 x_vertices, y_vertices, cell_to_vertex):
        """
        see @ref yac_cdef_grid_unstruct
        """
        cdef int grid_id
        cdef int[::1] num_vertices_per_cell_view = np.ascontiguousarray(num_vertices_per_cell, dtype=np.intc)
        cdef double[::1] x_vertices_view = np.ascontiguousarray(x_vertices, dtype=np.double)
        cdef double[::1] y_vertices_view = np.ascontiguousarray(y_vertices, dtype=np.double)
        cdef int[::1] cell_to_vertex_view = np.ascontiguousarray(cell_to_vertex, dtype=np.intc)
        yac_cdef_grid_unstruct(grid_name.encode(), len(x_vertices_view),
                               len(num_vertices_per_cell_view),
                               &num_vertices_per_cell_view[0],
                               &x_vertices_view[0], &y_vertices_view[0],
                               &cell_to_vertex_view[0], &grid_id)
        super().__init__(grid_id)

    def def_points(self, location : Location,
                   x_points, y_points):
        """
        see @ref yac_cdef_points_unstruct
        """
        cdef int points_id
        cdef double[::1] x_points_view = np.ascontiguousarray(x_points, dtype=np.double)
        cdef double[::1] y_points_view = np.ascontiguousarray(y_points, dtype=np.double)
        yac_cdef_points_unstruct(self.grid_id, len(x_points_view), location.value,
                                 &x_points_view[0], &y_points_view[0], &points_id)
        return Points(points_id)

class Field:
    """
    Stored the field_id
    """
    def __init__(self, field_id, size=None):
        self.field_id = field_id
        self._size = size

    @classmethod
    def create(cls, field_name : str, comp : Component, points):
        """
        see @ref yac_cdef_field
        """
        from collections.abc import Iterable
        cdef int field_id
        if not isinstance(points, Iterable):
            points = [points]
        cdef int[:] point_ids_array = np.array([p.points_id for p in points], dtype=np.intc)
        size = points[0].size
        yac_cdef_field(field_name.encode(), comp.comp_id,
                       &point_ids_array[0], len(point_ids_array), &field_id)
        return Field(field_id, size)

    def get(self, buf = None):
        """
        see @ref yac_cget_

        @param[out] buf    receive buffer, if `None` a numpy array of correct size is allocated
        """
        cdef int info
        cdef int ierror
        if buf is None:
            buf = np.empty((self.collection_size, self.size), dtype=np.double)
        buf = np.ascontiguousarray(buf.reshape(self.collection_size, self.size), dtype=np.double)
        cdef double[:,::1] buf_view = buf
        yac_cget_(self.field_id, buf_view.shape[0], &buf_view[0,0], &info, &ierror)
        if ierror != 0:
            raise RuntimeError("yac_cget returned error number " + str(ierror))
        return buf, info

    def put(self, buf):
        """
        see @ref yac_cput_
        """
        cdef int info
        cdef int ierror
        cdef double[:,:,::1] buf_view = np.ascontiguousarray(buf.reshape(-1,self.collection_size,self.size), dtype=np.double)
        yac_cput_(self.field_id, buf_view.shape[1], &buf_view[0,0,0], &info, &ierror)
        if ierror != 0:
            raise RuntimeError("yac_cput returned error number " + str(ierror))
        return info

    @property
    def name(self):
        """
        see @ref yac_cget_field_name_from_id
        """
        cdef char[64] buf
        yac_cget_field_name_from_id ( self.field_id, &buf[0] )
        return bytes.decode(buf)

    @property
    def role(self):
        """
        see @ref yac_cget_role_of_field_by_id
        """
        cdef int role
        yac_cget_role_of_field_by_id ( self.field_id, &role )
        return Role(role)

    @property
    def coupling_timestep(self):
        """
        see @ref yac_cget_coupling_timestep_id
        """
        cdef char[32] buf
        yac_cget_coupling_timestep_id ( self.field_id, &buf[0] )
        return bytes.decode(buf)

    @property
    def model_timestep(self):
        """
        see @ref yac_cget_model_timestep_id
        """
        cdef char[32] buf
        yac_cget_model_timestep_id ( self.field_id, &buf[0] )
        return bytes.decode(buf)

    @property
    def collection_size(self):
        """
        see @ref yac_cget_collection_size_id
        """
        cdef int collection_size
        yac_cget_collection_size_id(self.field_id, &collection_size)
        return collection_size

    @property
    def size(self):
        """
        The size of the corresponding points object
        """
        return self._size
        

def version():
    """
    see @ref yac_cget_version
    """
    return bytes.decode(yac_cget_version())
