/**
 * @file couple_config.c
 *
 * @copyright Copyright  (C)  2021 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "yac_mpi.h"
#include "couple_config.h"
#include "yac_interface.h"
#include "interp_method_avg.h"
#include "interp_method_check.h"
#include "interp_method_conserv.h"
#include "interp_method_creep.h"
#include "interp_method_nnn.h"
#include "interp_method_fixed.h"
#include "interp_method_file.h"
#include "interp_method_hcsbb.h"
#include "interp_method_spmap.h"
#include "interp_method_callback.h"

struct yac_couple_config * yac_couple_config_new() {

  struct yac_couple_config * couple_config =
    xmalloc(1 * sizeof(*couple_config));

  couple_config->timestep_unit = C_SECOND;
  couple_config->start_datetime = NULL;
  couple_config->end_datetime = NULL;
  couple_config->calendar = CALENDAR_NOT_SET;

  couple_config->grids = NULL;
  couple_config->num_grids = 0;

  couple_config->components = NULL;
  couple_config->num_components = 0;

  couple_config->transients = NULL;
  couple_config->num_transients = 0;

  couple_config->couples = NULL;
  couple_config->num_couples = 0;

  couple_config->redirect_stdout = 0;

  return couple_config;
}

static char const * string_dup(char const * string) {
  return (string != NULL)?strdup(string):NULL;
}

static struct yac_couple_config_grid *
  yac_couple_config_grids_copy(
    struct yac_couple_config_grid * grids, size_t num_grids) {

  struct yac_couple_config_grid * grids_copy =
    xmalloc(num_grids * sizeof(*grids_copy));

  for (size_t i = 0; i < num_grids; ++i) {
    grids_copy[i].name = string_dup(grids[i].name);
    grids_copy[i].id = grids[i].id;
  }

  return grids_copy;
}

static struct yac_couple_config_component_transient *
  yac_couple_config_component_transient_copy(
    struct yac_couple_config_component_transient * component_transients,
    size_t num_component_transients) {

  struct yac_couple_config_component_transient * component_transients_copy =
    xmalloc(num_component_transients * sizeof(*component_transients_copy));

  for (size_t i = 0; i < num_component_transients; ++i) {
    component_transients_copy[i].transient_idx =
      component_transients[i].transient_idx;
    component_transients_copy[i].grid_idx =
      component_transients[i].grid_idx;
  }

  return component_transients_copy;
}

static struct yac_couple_config_component *
  yac_couple_config_components_copy(
    struct yac_couple_config_component * components, size_t num_components) {

  struct yac_couple_config_component * components_copy =
    xmalloc(num_components * sizeof(*components_copy));

  for (size_t i = 0; i < num_components; ++i) {
    components_copy[i].component_transients =
      yac_couple_config_component_transient_copy(
        components[i].component_transients,
        components[i].num_component_transients);
    components_copy[i].num_component_transients =
      components[i].num_component_transients;
    components_copy[i].name = string_dup(components[i].name);
    components_copy[i].id = components[i].id;
  }

  return components_copy;
}

static struct yac_couple_config_transient *
  yac_couple_config_transients_copy(
    struct yac_couple_config_transient * transients, size_t num_transients) {

  struct yac_couple_config_transient * transients_copy =
    xmalloc(num_transients * sizeof(*transients_copy));

  for (size_t i = 0; i < num_transients; ++i) {
    transients_copy[i].name = string_dup(transients[i].name);
    transients_copy[i].id = transients[i].id;
    transients_copy[i].collection_size = transients[i].collection_size;
  }

  return transients_copy;
}

static struct yac_interp_stack_config copy_interp_stack(
  struct yac_interp_stack_config interp_stack) {

  struct yac_interp_stack_config interp_stack_copy;
  interp_stack_copy.size = interp_stack.size;
  interp_stack_copy.config =
    xmalloc(interp_stack_copy.size * sizeof(*(interp_stack_copy.config)));
  memcpy(
    interp_stack_copy.config, interp_stack.config,
    interp_stack_copy.size * sizeof(*(interp_stack_copy.config)));
  return interp_stack_copy;
}

static struct yac_couple_config_transient_couple *
  yac_couple_config_transient_couples_copy(
    struct yac_couple_config_transient_couple * transient_couples,
    size_t num_transient_couples) {

  struct yac_couple_config_transient_couple * transient_couples_copy =
    xmalloc(num_transient_couples * sizeof(*transient_couples_copy));

  for (size_t i = 0; i < num_transient_couples; ++i) {
    transient_couples_copy[i].source.component_idx =
      transient_couples[i].source.component_idx;
    transient_couples_copy[i].source.component_transient_idx =
      transient_couples[i].source.component_transient_idx;
    transient_couples_copy[i].source.timestep =
      string_dup(transient_couples[i].source.timestep);
    transient_couples_copy[i].source.lag =
      transient_couples[i].source.lag;

    transient_couples_copy[i].target.component_idx =
      transient_couples[i].target.component_idx;
    transient_couples_copy[i].target.component_transient_idx =
      transient_couples[i].target.component_transient_idx;
    transient_couples_copy[i].target.timestep =
      string_dup(transient_couples[i].target.timestep);
    transient_couples_copy[i].target.lag =
      transient_couples[i].target.lag;

    transient_couples_copy[i].mapping_on_source =
      transient_couples[i].mapping_on_source;

    transient_couples_copy[i].interp_stack =
      copy_interp_stack(transient_couples[i].interp_stack);

    transient_couples_copy[i].coupling_period_operation =
      transient_couples[i].coupling_period_operation;
    transient_couples_copy[i].coupling_period =
      string_dup(transient_couples[i].coupling_period);

    transient_couples_copy[i].enforce_write_weight_file =
      transient_couples[i].enforce_write_weight_file;
    transient_couples_copy[i].weight_file_name =
      string_dup(transient_couples[i].weight_file_name);
  }

  return transient_couples_copy;
}

static struct yac_couple_config_couple *
  yac_couple_config_couples_copy(
    struct yac_couple_config_couple * couples, size_t num_couples) {

  struct yac_couple_config_couple * couples_copy =
    xmalloc(num_couples * sizeof(*couples_copy));

  for (size_t i = 0; i < num_couples; ++i) {
    couples_copy[i].component_indices[0] = couples[i].component_indices[0];
    couples_copy[i].component_indices[1] = couples[i].component_indices[1];
    couples_copy[i].transient_couples =
      yac_couple_config_transient_couples_copy(
        couples[i].transient_couples, couples[i].num_transient_couples);
    couples_copy[i].num_transient_couples =
      couples[i].num_transient_couples;
  }

  return couples_copy;
}

struct yac_couple_config * yac_couple_config_copy(
  struct yac_couple_config * couple_config) {

  struct yac_couple_config * couple_config_copy =
    xmalloc(1 * sizeof(*couple_config));

  couple_config_copy->timestep_unit = couple_config->timestep_unit;
  couple_config_copy->start_datetime =
    string_dup(couple_config->start_datetime);
  couple_config_copy->end_datetime = string_dup(couple_config->end_datetime);
  couple_config_copy->calendar = couple_config->calendar;

  couple_config_copy->grids =
    yac_couple_config_grids_copy(
      couple_config->grids, couple_config->num_grids);
  couple_config_copy->num_grids = couple_config->num_grids;

  couple_config_copy->components =
    yac_couple_config_components_copy(
      couple_config->components, couple_config->num_components);
  couple_config_copy->num_components = couple_config->num_components;

  couple_config_copy->transients =
    yac_couple_config_transients_copy(
      couple_config->transients, couple_config->num_transients);
  couple_config_copy->num_transients = couple_config->num_transients;

  couple_config_copy->couples =
    yac_couple_config_couples_copy(
      couple_config->couples, couple_config->num_couples);
  couple_config_copy->num_couples = couple_config->num_couples;

  couple_config_copy->redirect_stdout = couple_config->redirect_stdout;

  return couple_config_copy;
}

void yac_couple_config_delete(struct yac_couple_config * couple_config) {

  free((void*)couple_config->start_datetime);
  free((void*)couple_config->end_datetime);

  for (size_t grid_idx = 0; grid_idx < couple_config->num_grids; ++grid_idx)
    free((void*)couple_config->grids[grid_idx].name);
  free(couple_config->grids);

  for (size_t component_idx = 0;
       component_idx < couple_config->num_components; ++component_idx) {
    free(couple_config->components[component_idx].component_transients);
    free((void*)couple_config->components[component_idx].name);
  }
  free(couple_config->components);

  for (size_t transient_idx = 0;
       transient_idx < couple_config->num_transients; ++transient_idx)
    free((void*)couple_config->transients[transient_idx].name);
  free(couple_config->transients);

  for (size_t couple_idx = 0; couple_idx < couple_config->num_couples;
       ++couple_idx) {
    for (size_t transient_couple_idx = 0;
         transient_couple_idx <
           couple_config->couples[couple_idx].num_transient_couples;
         ++transient_couple_idx) {

      free(
        (void*)couple_config->
          couples[couple_idx].
            transient_couples[transient_couple_idx].
              source.timestep);
      free(
        (void*)couple_config->
          couples[couple_idx].
            transient_couples[transient_couple_idx].
              target.timestep);

      free(
        couple_config->
          couples[couple_idx].
            transient_couples[transient_couple_idx].
              interp_stack.config);

      free(
        (void*)couple_config->
          couples[couple_idx].
            transient_couples[transient_couple_idx].
              coupling_period);

      free(
        (void*)couple_config->
          couples[couple_idx].
            transient_couples[transient_couple_idx].
              weight_file_name);
    }
    free(
      (void*)couple_config->
        couples[couple_idx].
          transient_couples);
  }
  free(couple_config->couples);
  free(couple_config);
}

void yac_couple_config_set_timestep_unit(
  struct yac_couple_config * couple_config,
  enum yac_time_unit_type timestep_unit) {

  YAC_ASSERT(
    (timestep_unit == C_MILLISECOND) ||
    (timestep_unit == C_SECOND) ||
    (timestep_unit == C_MINUTE) ||
    (timestep_unit == C_HOUR) ||
    (timestep_unit == C_DAY) ||
    (timestep_unit == C_MONTH) ||
    (timestep_unit == C_YEAR) ||
    (timestep_unit == C_ISO_FORMAT),
    "ERROR(yac_couple_config_set_timestep_unit): invalid timestep_unit");

  couple_config->timestep_unit = timestep_unit;
}

void yac_couple_config_set_calendar(
  struct yac_couple_config * couple_config, calendarType calendar) {

  YAC_ASSERT(
    (calendar == PROLEPTIC_GREGORIAN) ||
    (calendar == YEAR_OF_360_DAYS) ||
    (calendar == YEAR_OF_365_DAYS),
    "ERROR(yac_couple_config_set_calendar): invalid calendar")

  couple_config->calendar = calendar;
}

void yac_couple_config_add_grid(
  struct yac_couple_config * couple_config, char const * name, int id) {

  YAC_ASSERT(
    name != NULL,
    "ERROR(yac_couple_config_add_grid): invalid name (NULL is not allowed)")

  for (size_t i = 0; i < couple_config->num_grids; ++i)
    YAC_ASSERT(
      (couple_config->grids[i].id != id) &&
      strcmp(couple_config->grids[i].name, name),
      "ERROR(yac_couple_config_add_grid): duplicated grid id or name")

  size_t grid_idx = couple_config->num_grids;
  couple_config->num_grids++;
  couple_config->grids =
    xrealloc(
      couple_config->grids,
      couple_config->num_grids * sizeof(*(couple_config->grids)));
  couple_config->grids[grid_idx].name = string_dup(name);
  couple_config->grids[grid_idx].id = id;
}

void yac_couple_config_add_component(
  struct yac_couple_config * couple_config, char const * name, int id) {

  YAC_ASSERT(
    name != NULL,
    "ERROR(yac_couple_config_add_component): "
    "invalid name (NULL is not allowed)")

  for (size_t i = 0; i < couple_config->num_components; ++i)
    YAC_ASSERT(
      (couple_config->components[i].id != id) &&
      strcmp(couple_config->components[i].name, name),
      "ERROR(yac_couple_config_add_component): "
      "duplicated component id or name")

  size_t component_idx = couple_config->num_components;
  couple_config->num_components++;
  couple_config->components =
    xrealloc(
      couple_config->components,
      couple_config->num_components * sizeof(*(couple_config->components)));
  couple_config->components[component_idx].component_transients = NULL;
  couple_config->components[component_idx].num_component_transients = 0;
  couple_config->components[component_idx].name = strdup(name);
  couple_config->components[component_idx].id = id;
}

void yac_couple_config_add_transient(
  struct yac_couple_config * couple_config,
  char const * name, int id, size_t collection_size) {

  YAC_ASSERT(
    name != NULL,
    "ERROR(yac_couple_config_add_transient): "
    "invalid name (NULL is not allowed)")

  for (size_t i = 0; i < couple_config->num_transients; ++i)
    YAC_ASSERT(
      (couple_config->transients[i].id != id) &&
      strcmp(couple_config->transients[i].name, name),
      "ERROR(yac_couple_config_add_transient): "
      "duplicated transient id or name")

  size_t transient_idx = couple_config->num_transients;
  couple_config->num_transients++;
  couple_config->transients =
    xrealloc(
      couple_config->transients,
      couple_config->num_transients * sizeof(*(couple_config->transients)));
  couple_config->transients[transient_idx].name = strdup(name);
  couple_config->transients[transient_idx].id = id;
  couple_config->transients[transient_idx].collection_size = collection_size;
}

void yac_couple_config_set_transient_collection_size(
  struct yac_couple_config * couple_config,
  char const * name, size_t collection_size) {

  YAC_ASSERT(
    name != NULL,
    "ERROR(yac_couple_config_set_transient_collection_size): "
    "invalid name (NULL is not allowed)")

  size_t transient_idx;
  for (transient_idx = 0; transient_idx < couple_config->num_transients;
       ++transient_idx)
    if (!strcmp(couple_config->transients[transient_idx].name, name)) break;

  YAC_ASSERT(
    transient_idx != couple_config->num_transients,
    "ERROR(yac_couple_config_set_transient_collection_size): "
    "invalid transient name")

  couple_config->transients[transient_idx].collection_size = collection_size;
}

static size_t yac_couple_config_component_add_transient(
  struct yac_couple_config_component * component,
  size_t transient_idx, size_t grid_idx, int is_source) {

  for (size_t i = 0; i < component->num_component_transients; i++) {
    if ((component->component_transients[i].transient_idx == transient_idx) &&
        (component->component_transients[i].grid_idx == grid_idx)) {
          YAC_ASSERT(
            is_source,
            "ERROR(yac_couple_config_component_add_transient): "
            "duplicated component transient configuration");
          return i;
    }
  }

  size_t component_transient_idx = component->num_component_transients;
  component->num_component_transients++;

  component->component_transients =
    xrealloc(
      component->component_transients,
      component->num_component_transients *
      sizeof(*(component->component_transients)));

  component->component_transients[component_transient_idx].transient_idx =
    transient_idx;
  component->component_transients[component_transient_idx].grid_idx = grid_idx;

  return component_transient_idx;
}

static void yac_couple_config_transient_couple_init(
  struct yac_couple_config_transient_couple * transient_couple,
  struct yac_couple_config_component * components, size_t transient_idx,
  size_t src_component_idx, size_t src_grid_idx, char const * src_timestep, int src_lag,
  size_t tgt_component_idx, size_t tgt_grid_idx, char const * tgt_timestep, int tgt_lag,
  int mapping_on_source, struct yac_interp_stack_config interp_stack,
  enum yac_reduction_type coupling_period_operation,
  char const * coupling_period,
  int enforce_write_weight_file, char const * weight_file_name) {

  size_t src_component_transient_idx =
    yac_couple_config_component_add_transient(
      components + src_component_idx, transient_idx, src_grid_idx, 1);

  size_t tgt_component_transient_idx =
    yac_couple_config_component_add_transient(
      components + tgt_component_idx, transient_idx, tgt_grid_idx, 0);

  transient_couple->source.component_idx = src_component_idx;
  transient_couple->source.component_transient_idx =
    src_component_transient_idx;
  transient_couple->source.timestep = string_dup(src_timestep);
  transient_couple->source.lag = src_lag;

  transient_couple->target.component_idx = tgt_component_idx;
  transient_couple->target.component_transient_idx =
    tgt_component_transient_idx;
  transient_couple->target.timestep = string_dup(tgt_timestep);
  transient_couple->target.lag = tgt_lag;

  transient_couple->mapping_on_source = mapping_on_source;

  transient_couple->interp_stack = copy_interp_stack(interp_stack);

  transient_couple->coupling_period_operation = coupling_period_operation;
  transient_couple->coupling_period = string_dup(coupling_period);

  transient_couple->enforce_write_weight_file = enforce_write_weight_file;
  transient_couple->weight_file_name = string_dup(weight_file_name);
}

static size_t get_transient_idx(
  struct yac_couple_config_transient * transients, size_t num_transients,
  int transient_id) {

  size_t transient_idx = SIZE_MAX;
  for (size_t i = 0; (i < num_transients) && (transient_idx == SIZE_MAX); ++i)
    if (transients[i].id == transient_id) transient_idx = i;

  YAC_ASSERT(
    transient_idx != SIZE_MAX,
    "ERROR(get_transient_idx): transient id not in list of transients")

  return transient_idx;
}

static size_t get_component_idx(
  struct yac_couple_config_component * components, size_t num_components,
  int component_id) {

  size_t component_idx = SIZE_MAX;
  for (size_t i = 0; (i < num_components) && (component_idx == SIZE_MAX); ++i)
    if (components[i].id == component_id) component_idx = i;

  YAC_ASSERT(
    component_idx != SIZE_MAX,
    "ERROR(get_component_idx): component id not in list of components")

  return component_idx;
}

static size_t get_grid_idx(
  struct yac_couple_config_grid * grids, size_t num_grids,
  int grid_id) {

  size_t grid_idx = SIZE_MAX;
  for (size_t i = 0; (i < num_grids) && (grid_idx == SIZE_MAX); ++i)
    if (grids[i].id == grid_id) grid_idx = i;

  YAC_ASSERT(
    grid_idx != SIZE_MAX,
    "ERROR(get_grid_idx): grid id not in list of grids")

  return grid_idx;
}

void yac_couple_config_add_transient_couple(
  struct yac_couple_config * couple_config, int transient_id,
  int src_component_id, int src_grid_id, char const * src_timestep, int src_lag,
  int tgt_component_id, int tgt_grid_id, char const * tgt_timestep, int tgt_lag,
  int mapping_on_source, struct yac_interp_stack_config interp_stack,
  enum yac_reduction_type coupling_period_operation,
  char const * coupling_period,
  int enforce_write_weight_file, char const * weight_file_name) {

  size_t transient_idx =
    get_transient_idx(
      couple_config->transients, couple_config->num_transients, transient_id);

  size_t src_component_idx =
    get_component_idx(
      couple_config->components, couple_config->num_components,
      src_component_id);
  size_t tgt_component_idx =
    get_component_idx(
      couple_config->components, couple_config->num_components,
      tgt_component_id);
  size_t component_indices[2] = {src_component_idx, tgt_component_idx};
  if (src_component_idx > tgt_component_idx) {
    size_t temp_component_idx = component_indices[0];
    component_indices[0] = component_indices[1];
    component_indices[1] = temp_component_idx;
  }

  size_t src_grid_idx =
    get_grid_idx(couple_config->grids, couple_config->num_grids, src_grid_id);
  size_t tgt_grid_idx =
    get_grid_idx(couple_config->grids, couple_config->num_grids, tgt_grid_id);

  size_t couple_idx = 0;
  for (; couple_idx < couple_config->num_couples; ++couple_idx)
    if (!memcmp(
           couple_config->couples[couple_idx].component_indices,
           component_indices, sizeof(component_indices))) break;
  if (couple_idx == couple_config->num_couples) {
    couple_config->num_couples++;
    couple_config->couples =
      xrealloc(
        couple_config->couples,
        couple_config->num_couples * sizeof(*(couple_config->couples)));
    memcpy(couple_config->couples[couple_idx].component_indices,
           component_indices, sizeof(component_indices));
    couple_config->couples[couple_idx].transient_couples = NULL;
    couple_config->couples[couple_idx].num_transient_couples = 0;
  }

  struct yac_couple_config_couple * couple =
    couple_config->couples + couple_idx;
  couple->num_transient_couples++;
  couple->transient_couples =
    xrealloc(
      couple->transient_couples,
      couple->num_transient_couples * sizeof(*(couple->transient_couples)));

  struct yac_couple_config_transient_couple * transient_couple =
    couple->transient_couples + (couple->num_transient_couples - 1);

  yac_couple_config_transient_couple_init(
    transient_couple, couple_config->components, transient_idx,
    src_component_idx, src_grid_idx, src_timestep, src_lag,
    tgt_component_idx, tgt_grid_idx, tgt_timestep, tgt_lag,
    mapping_on_source, interp_stack, coupling_period_operation,
    coupling_period, enforce_write_weight_file, weight_file_name);
}

static int compare_int(const void * a, const void * b) {
  return *(int*)a - *(int*)b;
}

int yac_couple_config_get_free_grid_id(
  struct yac_couple_config * couple_config) {

  if (couple_config->num_grids == 0) return 1;

  int * all_grid_ids =
    xmalloc(couple_config->num_grids * sizeof(*all_grid_ids));

  for (size_t i = 0; i < couple_config->num_grids; ++i)
    all_grid_ids[i] = couple_config->grids[i].id;
  qsort(
    all_grid_ids, couple_config->num_grids, sizeof(*all_grid_ids), compare_int);

  int free_grid_id = all_grid_ids[0] + 1;

  for (size_t i = 1; i < couple_config->num_grids; ++i) {
    if (all_grid_ids[i] > free_grid_id) {
      break;
    } else {
      free_grid_id = all_grid_ids[i] + 1;
    }
  }

  free(all_grid_ids);

  return free_grid_id;
}

int yac_couple_config_get_free_component_id(
  struct yac_couple_config * couple_config) {

  if (couple_config->num_components == 0) return 1;

  int * all_component_ids =
    xmalloc(couple_config->num_components * sizeof(*all_component_ids));

  for (size_t i = 0; i < couple_config->num_components; ++i)
    all_component_ids[i] = couple_config->components[i].id;
  qsort(
    all_component_ids, couple_config->num_components,
    sizeof(*all_component_ids), compare_int);

  int free_component_id = all_component_ids[0] + 1;

  for (size_t i = 1; i < couple_config->num_components; ++i) {
    if (all_component_ids[i] > free_component_id) {
      break;
    } else {
      free_component_id = all_component_ids[i] + 1;
    }
  }

  free(all_component_ids);

  return free_component_id;
}

int yac_couple_config_get_free_transient_id(
  struct yac_couple_config * couple_config) {

  if (couple_config->num_transients == 0) return 1;

  int * all_transient_ids =
    xmalloc(couple_config->num_transients * sizeof(*all_transient_ids));

  for (size_t i = 0; i < couple_config->num_transients; ++i)
    all_transient_ids[i] = couple_config->transients[i].id;
  qsort(
    all_transient_ids, couple_config->num_transients,
    sizeof(*all_transient_ids), compare_int);

  int free_transient_id = all_transient_ids[0] + 1;

  for (size_t i = 1; i < couple_config->num_transients; ++i) {
    if (all_transient_ids[i] > free_transient_id) {
      break;
    } else {
      free_transient_id = all_transient_ids[i] + 1;
    }
  }

  free(all_transient_ids);

  return free_transient_id;
}

size_t yac_couple_config_get_num_couples(
  struct yac_couple_config * couple_config) {

  return couple_config->num_couples;
}

static void check_couple_idx(
  struct yac_couple_config * couple_config, size_t couple_idx,
  char const * routine_name, int line) {

  YAC_ASSERT_F(
    (couple_idx >= 0) && (couple_idx < couple_config->num_couples),
    "ERROR(%s:%d:%s): invalid couple_idx", __FILE__, line, routine_name);
}

size_t yac_couple_config_get_num_couple_transients(
  struct yac_couple_config * couple_config, size_t couple_idx) {

  check_couple_idx(
    couple_config, couple_idx, "yac_couple_config_get_num_couple_transients",
    __LINE__);

  return couple_config->couples[couple_idx].num_transient_couples;
}

void yac_couple_config_get_couple_component_ids(
  struct yac_couple_config * couple_config, size_t couple_idx,
  int couple_component_ids[2]) {

  check_couple_idx(
    couple_config, couple_idx, "yac_couple_config_get_couple_component_ids",
    __LINE__);

  for (int i = 0; i < 2; ++i)
    couple_component_ids[i] =
      couple_config->components[
        couple_config->couples[couple_idx].component_indices[i]].id;
}


int yac_couple_config_component_name_is_valid(
  struct yac_couple_config * couple_config, char const * component_name) {

  YAC_ASSERT(
    strlen(component_name) <= YAC_MAX_CHARLEN,
    "ERROR(yac_couple_config_component_name_is_valid): "
    "component name is too long (maximum is YAC_MAX_CHARLEN)")

  for (size_t component_idx = 0; component_idx < couple_config->num_components;
       ++component_idx)
    if (!strcmp(component_name, couple_config->components[component_idx].name))
      return 1;
  return 0;
}

size_t yac_couple_config_get_num_components(
  struct yac_couple_config * couple_config) {

  return couple_config->num_components;
}

static void check_component_idx(
  struct yac_couple_config * couple_config, size_t component_idx,
  char const * routine_name, int line) {

  YAC_ASSERT_F(
    (component_idx >= 0) && (component_idx < couple_config->num_components),
    "ERROR(%s:%d:%s): invalid component_idx", __FILE__, line, routine_name)
}

int yac_couple_config_get_component_id(
  struct yac_couple_config * couple_config, size_t component_idx) {

  check_component_idx(
    couple_config, component_idx,
    "yac_couple_config_get_component_id", __LINE__);

  return couple_config->components[component_idx].id;
}

char const * yac_couple_config_get_component_name(
  struct yac_couple_config * couple_config, size_t component_idx) {

  check_component_idx(
    couple_config, component_idx,
    "yac_couple_config_get_component_name", __LINE__);

  return couple_config->components[component_idx].name;
}

char const * yac_couple_config_get_component_name_by_id(
  struct yac_couple_config * couple_config, int component_id) {

  char const * comp_name = NULL;
  for (size_t component_idx = 0;
       (component_idx < couple_config->num_components) && (comp_name == NULL);
       ++component_idx)
    if (couple_config->components[component_idx].id == component_id)
      comp_name = couple_config->components[component_idx].name;

  YAC_ASSERT_F(
    comp_name != NULL,
    "ERROR(yac_couple_config_get_component_name_by_id): "
    "No component with ID %d found.", component_id)
  return comp_name;
}

int yac_couple_config_get_component_id_by_name(
  struct yac_couple_config * couple_config, char const * component_name) {

  int comp_id = INT_MAX;
  for (size_t component_idx = 0;
       (component_idx < couple_config->num_components) && (comp_id == INT_MAX);
       ++component_idx)
    if (!strcmp(couple_config->components[component_idx].name, component_name))
      comp_id = couple_config->components[component_idx].id;

  YAC_ASSERT_F(
    comp_id != INT_MAX,
    "ERROR(yac_couple_config_get_component_id_by_name): "
    "No component with name \"%s\" found.", component_name)
  return comp_id;
}

static void check_transient_couple_idx(
  struct yac_couple_config * couple_config, size_t couple_idx,
  size_t transient_couple_idx, char const * routine_name, int line) {

  check_couple_idx(couple_config, couple_idx, routine_name, __LINE__);

  YAC_ASSERT_F(
    (transient_couple_idx >= 0) &&
    (transient_couple_idx <
       couple_config->couples[couple_idx].num_transient_couples),
    "ERROR(%s:%d:%s): invalid transient_couple_idx",
    __FILE__, line, routine_name)
}

struct yac_interp_stack_config yac_couple_config_get_interp_stack(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_get_interp_stack", __LINE__);

  return
    couple_config->
      couples[couple_idx].
        transient_couples[transient_couple_idx].interp_stack;
}

void yac_couple_config_get_transient_grid_names(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx,
  char const ** src_grid_name, char const ** tgt_grid_name) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_get_transient_grid_names", __LINE__);

  size_t src_component_idx =
    couple_config->couples[couple_idx].
      transient_couples[transient_couple_idx].
        source.component_idx;
  size_t src_component_transient_idx =
    couple_config->couples[couple_idx].
      transient_couples[transient_couple_idx].
        source.component_transient_idx;

  size_t tgt_component_idx =
    couple_config->couples[couple_idx].
      transient_couples[transient_couple_idx].
        target.component_idx;
  size_t tgt_component_transient_idx =
    couple_config->couples[couple_idx].
      transient_couples[transient_couple_idx].
        target.component_transient_idx;

  *src_grid_name =
    couple_config->grids[
      couple_config->components[src_component_idx].
        component_transients[src_component_transient_idx].grid_idx].name;
  *tgt_grid_name =
    couple_config->grids[
      couple_config->components[tgt_component_idx].
        component_transients[tgt_component_transient_idx].grid_idx].name;
}


size_t yac_couple_config_get_transient_collection_size(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_get_transient_collection_size", __LINE__);

  size_t component_idx =
    couple_config->couples[couple_idx].
      transient_couples[transient_couple_idx].source.component_idx;
  size_t component_transient_idx =
    couple_config->couples[couple_idx].
      transient_couples[transient_couple_idx].source.component_transient_idx;

  return
    couple_config->transients[
      couple_config->components[component_idx].
        component_transients[component_transient_idx].
          transient_idx].collection_size;
}

void yac_couple_config_get_transient_couple_component_ids(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx,
  int * src_component_id, int * tgt_component_id) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_get_transient_couple_component_ids", __LINE__);

  *src_component_id =
    couple_config->components[
      couple_config->couples[couple_idx].
        transient_couples[transient_couple_idx].source.component_idx].id;
  *tgt_component_id =
    couple_config->components[
      couple_config->couples[couple_idx].
        transient_couples[transient_couple_idx].target.component_idx].id;
}

char const * yac_couple_config_get_transient_name(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_get_transient_name", __LINE__);

  size_t component_idx =
    couple_config->couples[couple_idx].
      transient_couples[transient_couple_idx].source.component_idx;
  size_t component_transient_idx =
    couple_config->couples[couple_idx].
      transient_couples[transient_couple_idx].source.component_transient_idx;

  return
    couple_config->transients[
      couple_config->components[component_idx].
        component_transients[component_transient_idx].
          transient_idx].name;
}

int yac_couple_config_mapping_on_source(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_mapping_on_source", __LINE__);

  return
    couple_config->
      couples[couple_idx].
        transient_couples[transient_couple_idx].
          mapping_on_source;
}

enum yac_time_unit_type yac_couple_config_get_timestep_unit(
  struct yac_couple_config * couple_config) {

  return couple_config->timestep_unit;
}

char const * yac_couple_config_get_coupling_period(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_get_coupling_period", __LINE__);

  return
    couple_config->
      couples[couple_idx].
        transient_couples[transient_couple_idx].
          coupling_period;
}

void yac_couple_config_get_source_timestep(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx,
  char const ** timestep, int * lag) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_get_source_timestep", __LINE__);

  *timestep =
    couple_config->
      couples[couple_idx].
        transient_couples[transient_couple_idx].
          source.timestep;
  *lag =
    couple_config->
      couples[couple_idx].
        transient_couples[transient_couple_idx].
          source.lag;
}

void yac_couple_config_get_target_timestep(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx,
  char const ** timestep, int * lag) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_get_target_timestep", __LINE__);

  *timestep =
    couple_config->
      couples[couple_idx].
        transient_couples[transient_couple_idx].
          target.timestep;
  *lag =
    couple_config->
      couples[couple_idx].
        transient_couples[transient_couple_idx].
          target.lag;
}

enum yac_reduction_type yac_couple_config_get_coupling_period_operation(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_get_coupling_period_operation", __LINE__);

  return
    couple_config->
      couples[couple_idx].
        transient_couples[transient_couple_idx].
          coupling_period_operation;
}

void yac_couple_config_set_datetime(
  struct yac_couple_config * couple_config,
  char const * start, char const * end) {

  if ((start != NULL) && (strlen(start) > 0)) {

    free((void*)couple_config->start_datetime);
    couple_config->start_datetime = strdup(start);
  }

  if ((end != NULL) && (strlen(end) > 0)) {

    free((void*)couple_config->end_datetime);
    couple_config->end_datetime = strdup(end);
  }
}

char const * yac_couple_config_get_start_datetime(
  struct yac_couple_config * couple_config) {

  return couple_config->start_datetime;
}

char const * yac_couple_config_get_end_datetime(
  struct yac_couple_config * couple_config) {

  return couple_config->end_datetime;
}

int yac_couple_config_get_grid_id_by_name(
  struct yac_couple_config * couple_config, char const * grid_name) {

  int grid_id = INT_MAX;
  for (size_t grid_idx = 0;
      (grid_idx < couple_config->num_grids) && (grid_id == INT_MAX); ++grid_idx)
    if (!strcmp(couple_config->grids[grid_idx].name, grid_name))
      grid_id = couple_config->grids[grid_idx].id;

  YAC_ASSERT_F(
    grid_id != INT_MAX,
    "ERROR(yac_couple_config_get_grid_id_by_name): "
    "No grid with name \"%s\" found.", grid_name)
  return grid_id;
}

int yac_couple_config_enforce_write_weight_file(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_enforce_write_weight_file", __LINE__);

  return
    couple_config->
      couples[couple_idx].
        transient_couples[transient_couple_idx].
          enforce_write_weight_file;
}

char const * yac_couple_config_get_weight_file_name(
  struct yac_couple_config * couple_config,
  size_t couple_idx, size_t transient_couple_idx) {

  check_transient_couple_idx(
    couple_config, couple_idx, transient_couple_idx,
    "yac_couple_config_get_weight_file_name", __LINE__);

  static char dummy[] = "\0";
  char const * weight_file_name =
    couple_config->
      couples[couple_idx].
        transient_couples[transient_couple_idx].
          weight_file_name;

  return (weight_file_name != NULL)?weight_file_name:dummy;
}

int yac_couple_config_redirect_stdout(
  struct yac_couple_config * couple_config) {

  return couple_config->redirect_stdout;
}

int yac_couple_config_contains_grid_name(
  struct yac_couple_config * couple_config, char const * grid_name) {

  for (size_t grid_idx = 0; grid_idx < couple_config->num_grids;
       ++grid_idx)
    if (!strcmp(couple_config->grids[grid_idx].name, grid_name))
      return 1;
  return 0;
}

int yac_couple_config_get_transient_id_by_name(
  struct yac_couple_config * couple_config, char const * transient_name) {

  int transient_id = INT_MAX;
  for (size_t transient_idx = 0;
       (transient_idx < couple_config->num_transients) &&
       (transient_id == INT_MAX); ++transient_idx)
    if (!strcmp(couple_config->transients[transient_idx].name, transient_name))
      transient_id = couple_config->transients[transient_idx].id;

  YAC_ASSERT_F(
    transient_id != INT_MAX,
    "ERROR(yac_couple_config_get_transient_id_by_name): "
    "No transient with name \"%s\" found.", transient_name)
  return transient_id;
}

int yac_couple_config_transient_config_is_valid(
  struct yac_couple_config * couple_config,
  int component_id, int grid_id, int transient_id) {

  size_t component_idx;
  for (component_idx = 0; component_idx < couple_config->num_components;
       ++component_idx)
    if (couple_config->components[component_idx].id == component_id) break;
  if (component_idx >= couple_config->num_components) return 0;

  size_t grid_idx;
  for (grid_idx = 0; grid_idx < couple_config->num_grids;
       ++grid_idx)
    if (couple_config->grids[grid_idx].id == grid_id) break;
  if (grid_idx >= couple_config->num_grids) return 0;

  size_t transient_idx;
  for (transient_idx = 0; transient_idx < couple_config->num_transients;
       ++transient_idx)
    if (couple_config->transients[transient_idx].id == transient_id) break;
  if (transient_idx >= couple_config->num_transients) return 0;

  for (size_t component_transient_idx = 0;
       component_transient_idx <
         couple_config->components[component_idx].num_component_transients;
       ++component_transient_idx)
    if ((couple_config->
           components[component_idx].
             component_transients[component_transient_idx].
               transient_idx == transient_idx) &&
        (couple_config->
           components[component_idx].
             component_transients[component_transient_idx].
               grid_idx == grid_idx)) return 1;
  return 0;
}

calendarType yac_couple_config_get_calendar(
  struct yac_couple_config * couple_config) {

  return couple_config->calendar;
}

static void inline check_interpolation_type(
  enum yac_interpolation_list type, char const * routine) {

  YAC_ASSERT_F(
    (type == AVERAGE) ||
    (type == RADIAL_BASIS_FUNCTION) ||
    (type == N_NEAREST_NEIGHBOR) ||
    (type == CONSERVATIVE) ||
    (type == SOURCE_TO_TARGET_MAP) ||
    (type == FIXED_VALUE) ||
    (type == BERNSTEIN_BEZIER) ||
    (type == USER_FILE) ||
    (type == CHECK) ||
    (type == CREEP) ||
    (type == USER_CALLBACK) ||
    (type == UNDEFINED),
    "ERROR(%s): invalid interpolation type", routine)
}

int yac_interp_stack_config_entry_compare(void const * a_, void const * b_) {

  union yac_interp_stack_config_entry const * a =
   (union yac_interp_stack_config_entry const *)a_;
  union yac_interp_stack_config_entry const * b =
   (union yac_interp_stack_config_entry const *)b_;

  check_interpolation_type(
    a->general.type, "yac_interp_stack_config_entry_compare");
  check_interpolation_type(
    b->general.type, "yac_interp_stack_config_entry_compare");

  if (a->general.type != b->general.type)
    return (a->general.type > b->general.type) -
           (a->general.type < b->general.type);

  YAC_ASSERT(
    (a->general.type == AVERAGE) ||
    (a->general.type == N_NEAREST_NEIGHBOR) ||
    (a->general.type == CONSERVATIVE) ||
    (a->general.type == SOURCE_TO_TARGET_MAP) ||
    (a->general.type == FIXED_VALUE) ||
    (a->general.type == USER_FILE) ||
    (a->general.type == CHECK) ||
    (a->general.type == BERNSTEIN_BEZIER) ||
    (a->general.type == RADIAL_BASIS_FUNCTION) ||
    (a->general.type == CREEP) ||
    (a->general.type == USER_CALLBACK),
    "ERROR(yac_interp_stack_config_entry_compare): invalid interpolation type")

  switch(a->general.type) {
    default:
    case (BERNSTEIN_BEZIER):
      return 0;
    case (AVERAGE): {
      if (a->average.reduction_type != b->average.reduction_type)
        return (a->average.reduction_type > b->average.reduction_type) -
               (a->average.reduction_type < b->average.reduction_type);
      return (a->average.partial_coverage > b->average.partial_coverage) -
             (a->average.partial_coverage < b->average.partial_coverage);
    }
    case (RADIAL_BASIS_FUNCTION):
    case (N_NEAREST_NEIGHBOR): {
      if (a->n_nearest_neighbor.config.n !=
          b->n_nearest_neighbor.config.n)
        return (a->n_nearest_neighbor.config.n >
                b->n_nearest_neighbor.config.n) -
               (a->n_nearest_neighbor.config.n <
                b->n_nearest_neighbor.config.n);
      if (a->n_nearest_neighbor.config.type !=
          b->n_nearest_neighbor.config.type)
        return (a->n_nearest_neighbor.config.type >
                b->n_nearest_neighbor.config.type) -
               (a->n_nearest_neighbor.config.type <
                b->n_nearest_neighbor.config.type);
      switch (a->n_nearest_neighbor.config.type) {
        case (NNN_GAUSS):
          return (a->n_nearest_neighbor.config.data.gauss_scale >
                  b->n_nearest_neighbor.config.data.gauss_scale) -
                 (a->n_nearest_neighbor.config.data.gauss_scale <
                  b->n_nearest_neighbor.config.data.gauss_scale);
        case (NNN_RBF):
          return (a->n_nearest_neighbor.config.data.rbf_scale >
                  b->n_nearest_neighbor.config.data.rbf_scale) -
                 (a->n_nearest_neighbor.config.data.rbf_scale <
                  b->n_nearest_neighbor.config.data.rbf_scale);
        default:
          return 0;
      };
    }
    case (CONSERVATIVE): {
      if (a->conservative.order !=
          b->conservative.order)
        return (a->conservative.order >
                b->conservative.order) -
               (a->conservative.order <
                b->conservative.order);
      if (a->conservative.enforced_conserv !=
          b->conservative.enforced_conserv)
        return (a->conservative.enforced_conserv >
                b->conservative.enforced_conserv) -
               (a->conservative.enforced_conserv <
                b->conservative.enforced_conserv);
      if (a->conservative.partial_coverage !=
          b->conservative.partial_coverage)
        return (a->conservative.partial_coverage >
                b->conservative.partial_coverage) -
               (a->conservative.partial_coverage <
                b->conservative.partial_coverage);
      return (a->conservative.normalisation >
              b->conservative.normalisation) -
              (a->conservative.normalisation <
              b->conservative.normalisation);
    }
    case (SOURCE_TO_TARGET_MAP): {
      if (a->spmap.spread_distance !=
          b->spmap.spread_distance)
        return (a->spmap.spread_distance >
                b->spmap.spread_distance) -
               (a->spmap.spread_distance <
                b->spmap.spread_distance);
      if (a->spmap.max_search_distance !=
          b->spmap.max_search_distance)
        return (a->spmap.max_search_distance >
                b->spmap.max_search_distance) -
               (a->spmap.max_search_distance <
                b->spmap.max_search_distance);
      return (a->spmap.weight_type >
              b->spmap.weight_type) -
             (a->spmap.weight_type <
              b->spmap.weight_type);
    }
    case (FIXED_VALUE): {
      return (a->fixed.value >
              b->fixed.value) -
             (a->fixed.value <
              b->fixed.value);
    }
    case (USER_FILE): {
      int ret;
      if ((ret = strcmp(a->user_file.filename, b->user_file.filename)))
        return ret;
      if ((ret = strcmp(a->user_file.src_grid_name, b->user_file.src_grid_name)))
        return ret;
      if ((ret = strcmp(a->user_file.tgt_grid_name, b->user_file.tgt_grid_name)))
        return ret;
      return (a->user_file.io_proc_step >
              b->user_file.io_proc_step) -
             (a->user_file.io_proc_step <
              b->user_file.io_proc_step);
    }
    case (CHECK): {
      int ret;
      if ((ret = strncmp(a->check.constructor_key, b->check.constructor_key,
                         sizeof(a->check.constructor_key)))) return ret;
      return
        strncmp(
          a->check.do_search_key, b->check.do_search_key,
          sizeof(a->check.do_search_key));
    };
    case (CREEP): {
      return
        (a->creep.creep_distance > b->creep.creep_distance) -
        (a->creep.creep_distance < b->creep.creep_distance);
    }
    case (USER_CALLBACK): {
      return strcmp(a->user_callback.func_compute_weights_key,
                    b->user_callback.func_compute_weights_key);
    }
  };
}

int yac_interp_stack_config_compare(void const * a_, void const * b_) {

  struct yac_interp_stack_config const * a =
   (struct yac_interp_stack_config const *)a_;
  struct yac_interp_stack_config const * b =
   (struct yac_interp_stack_config const *)b_;

  int ret;
  if ((ret = (a->size > b->size) - (a->size < b->size))) return ret;

  size_t stack_size = a->size;
  for (size_t method_idx = 0; method_idx < stack_size; ++method_idx)
    if ((ret =
           yac_interp_stack_config_entry_compare(
             &(a->config[method_idx]), &(b->config[method_idx]))))
      return ret;
  return 0;
}

struct interp_method ** yac_interp_stack_config_generate(
  struct yac_interp_stack_config interp_stack) {

  size_t interp_stack_size = interp_stack.size;
  struct interp_method ** method_stack =
    xmalloc((interp_stack_size + 1) * sizeof(*method_stack));
  method_stack[interp_stack_size] = NULL;

  for (size_t i = 0; i < interp_stack_size; ++i) {

    check_interpolation_type(
      interp_stack.config[i].general.type, "yac_interp_stack_config_generate");
    YAC_ASSERT(
      interp_stack.config[i].general.type != UNDEFINED,
      "ERROR(yac_interp_stack_config_generate): "
      "unsupported interpolation method")
    switch((int)(interp_stack.config[i].general.type)) {
      default:
      case(AVERAGE): {
        enum yac_interp_avg_weight_type weight_type =
          interp_stack.config[i].average.reduction_type;
        int partial_coverage =
          (int)interp_stack.config[i].average.partial_coverage;
        method_stack[i] =
          yac_interp_method_avg_new(weight_type, partial_coverage);
        break;
      }
      case(CONSERVATIVE): {
        int order =
          interp_stack.config[i].conservative.order;
        int enforced_conserv =
          interp_stack.config[i].conservative.enforced_conserv;
        int partial_coverage =
          interp_stack.config[i].conservative.partial_coverage;
        enum yac_interp_method_conserv_normalisation normalisation =
          interp_stack.config[i].conservative.normalisation;

        method_stack[i] =
          yac_interp_method_conserv_new(
            order, enforced_conserv, partial_coverage, normalisation);
        break;
      }
      case(FIXED_VALUE): {
        double fixed_value =
          interp_stack.config[i].fixed.value;
        method_stack[i] = yac_interp_method_fixed_new(fixed_value);
        break;
      }
      case(USER_FILE): {
        char const * weight_file_name =
          interp_stack.config[i].user_file.filename;
        int io_proc_step =
          interp_stack.config[i].user_file.io_proc_step;
        char const * src_grid_name =
          interp_stack.config[i].user_file.src_grid_name;
        char const * tgt_grid_name =
          interp_stack.config[i].user_file.tgt_grid_name;
        method_stack[i] =
          yac_interp_method_file_new(
            weight_file_name, io_proc_step, src_grid_name, tgt_grid_name);
        break;
      }
      case(CHECK): {
        func_constructor constructor_callback;
        void * constructor_user_data;
        func_do_search do_search_callback;
        void * do_search_user_data;

        yac_interp_method_check_get_constructor_callback(
          interp_stack.config[i].check.constructor_key,
          &constructor_callback, &constructor_user_data);
        yac_interp_method_check_get_do_search_callback(
          interp_stack.config[i].check.do_search_key,
          &do_search_callback, &do_search_user_data);

        method_stack[i] =
          yac_interp_method_check_new(
            constructor_callback, constructor_user_data,
            do_search_callback, do_search_user_data);
        break;
      }
      case (N_NEAREST_NEIGHBOR): {
        method_stack[i] =
          yac_interp_method_nnn_new(
            interp_stack.config[i].
              n_nearest_neighbor.config);
        break;
      }
      case (BERNSTEIN_BEZIER): {
        method_stack[i] =
          yac_interp_method_hcsbb_new();
        break;
      }
      case (RADIAL_BASIS_FUNCTION): {
        method_stack[i] =
          yac_interp_method_nnn_new(
            interp_stack.config[i].
              radial_basis_function.config);
        break;
      }
      case (SOURCE_TO_TARGET_MAP): {
        method_stack[i] =
          yac_interp_method_spmap_new(
            interp_stack.config[i].spmap.spread_distance,
            interp_stack.config[i].spmap.max_search_distance,
            interp_stack.config[i].spmap.weight_type);
        break;
      }
      case (CREEP): {
        method_stack[i] =
          yac_interp_method_creep_new(
            interp_stack.config[i].creep.creep_distance);
        break;
      }
      case(USER_CALLBACK): {
        yac_func_compute_weights compute_weights_callback;
        void * user_data;
        yac_interp_method_callback_get_compute_weights_callback(
          interp_stack.config[i].user_callback.func_compute_weights_key,
          &compute_weights_callback, &user_data);

        method_stack[i] =
          yac_interp_method_callback_new(compute_weights_callback, user_data);
        break;
      }
    };
  }

  return method_stack;
}

static size_t yac_couple_config_get_string_pack_size(
  char const * string, MPI_Comm comm) {

  int strlen_pack_size, string_pack_size;
  yac_mpi_call(MPI_Pack_size(1, MPI_INT, comm, &strlen_pack_size), comm);

  if (string != NULL) {
    yac_mpi_call(
      MPI_Pack_size(
        (int)(strlen(string)), MPI_CHAR, comm, &string_pack_size), comm);
  } else {
    string_pack_size = 0;
  }

  return (size_t)strlen_pack_size + (size_t)string_pack_size;
}

static size_t yac_couple_config_get_grid_pack_size(
  struct yac_couple_config_grid * grid, MPI_Comm comm) {

  int id_pack_size;
  yac_mpi_call(MPI_Pack_size(1, MPI_INT, comm, &id_pack_size), comm);

  return
    (size_t)id_pack_size +
    yac_couple_config_get_string_pack_size(grid->name, comm);
}

static size_t yac_couple_config_get_grids_pack_size(
  size_t num_grids, struct yac_couple_config_grid * grids, MPI_Comm comm) {

  int num_grids_pack_size;
  yac_mpi_call(MPI_Pack_size(1, MPI_INT, comm, &num_grids_pack_size), comm);

  size_t grids_pack_size = 0;
  for (size_t i = 0; i < num_grids; ++i)
    grids_pack_size +=
      yac_couple_config_get_grid_pack_size(grids + i, comm);

  return (size_t)num_grids_pack_size + grids_pack_size;
}

static size_t yac_couple_config_get_component_transient_pack_size(
  struct yac_couple_config_component_transient * component_transient,
  MPI_Comm comm) {

  YAC_ASSERT(
    (component_transient->transient_idx >= 0) &&
    (component_transient->transient_idx <= INT_MAX),
    "ERROR(yac_couple_config_get_component_transient_pack_size):"
    "transient_idx is too big")

  YAC_ASSERT(
    (component_transient->grid_idx >= 0) &&
    (component_transient->grid_idx <= INT_MAX),
    "ERROR(yac_couple_config_get_component_transient_pack_size):"
    "grid_idx is too big")

  int indices_pack_size;
  yac_mpi_call(MPI_Pack_size(2, MPI_INT, comm, &indices_pack_size), comm);

  return (size_t)indices_pack_size;
}

static size_t yac_couple_config_get_component_transients_pack_size(
  size_t num_component_transients,
  struct yac_couple_config_component_transient * component_transients,
  MPI_Comm comm) {

  int num_component_transients_pack_size;
  yac_mpi_call(
    MPI_Pack_size(1, MPI_INT, comm, &num_component_transients_pack_size), comm);

  size_t component_transients_pack_size = 0;
  for (size_t i = 0; i < num_component_transients; ++i)
    component_transients_pack_size +=
      yac_couple_config_get_component_transient_pack_size(
        component_transients + i, comm);

  return (size_t)num_component_transients_pack_size +
         component_transients_pack_size;
}

static size_t yac_couple_config_get_component_pack_size(
  struct yac_couple_config_component * component, MPI_Comm comm) {

  int id_pack_size;
  yac_mpi_call(MPI_Pack_size(1, MPI_INT, comm, &id_pack_size), comm);

  return
    (size_t)id_pack_size +
    yac_couple_config_get_string_pack_size(component->name, comm) +
    yac_couple_config_get_component_transients_pack_size(
      component->num_component_transients,
      component->component_transients, comm);
}

static size_t yac_couple_config_get_components_pack_size(
  size_t num_components, struct yac_couple_config_component * components,
  MPI_Comm comm) {

  int num_components_pack_size;
  yac_mpi_call(
    MPI_Pack_size(1, MPI_INT, comm, &num_components_pack_size), comm);

  size_t components_pack_size = 0;
  for (size_t i = 0; i < num_components; ++i)
    components_pack_size +=
      yac_couple_config_get_component_pack_size(components + i, comm);

  return (size_t)num_components_pack_size + components_pack_size;
}

static size_t yac_couple_config_get_transient_pack_size(
  struct yac_couple_config_transient * transient, MPI_Comm comm) {

  int ints_pack_size;
  yac_mpi_call(MPI_Pack_size(2, MPI_INT, comm, &ints_pack_size), comm);

  return
    (size_t)ints_pack_size + // id + collection_size
    yac_couple_config_get_string_pack_size(transient->name, comm);
}

static size_t yac_couple_config_get_transients_pack_size(
  size_t num_transients, struct yac_couple_config_transient * transients,
  MPI_Comm comm) {

  int num_transients_pack_size;
  yac_mpi_call(
    MPI_Pack_size(1, MPI_INT, comm, &num_transients_pack_size), comm);

  size_t transients_pack_size = 0;
  for (size_t i = 0; i < num_transients; ++i)
    transients_pack_size +=
      yac_couple_config_get_transient_pack_size(transients + i, comm);

  return (size_t)num_transients_pack_size + transients_pack_size;
}

static size_t yac_couple_config_get_interp_stack_entry_pack_size(
  union yac_interp_stack_config_entry * entry, MPI_Comm comm) {

  int int_pack_size, dbl_pack_size;
  yac_mpi_call(MPI_Pack_size(1, MPI_INT, comm, &int_pack_size), comm);
  yac_mpi_call(MPI_Pack_size(1, MPI_DOUBLE, comm, &dbl_pack_size), comm);

  check_interpolation_type(
    entry->general.type,
    "yac_couple_config_get_interp_stack_entry_pack_size");
  YAC_ASSERT(
    entry->general.type != UNDEFINED,
    "ERROR(yac_couple_config_get_interp_stack_entry_pack_size): "
    "invalid interpolation type")
  switch (entry->general.type) {
    default:
    case (AVERAGE):
      return (size_t)int_pack_size + // type
             (size_t)int_pack_size + // reduction_type
             (size_t)int_pack_size;  // partial_coverage
    case (RADIAL_BASIS_FUNCTION):
    case (N_NEAREST_NEIGHBOR):
      return (size_t)int_pack_size + // type
             (size_t)int_pack_size + // weight_type
             (size_t)int_pack_size + // n
             (size_t)dbl_pack_size;  // scale
    case (CONSERVATIVE):
      return (size_t)int_pack_size + // type
             (size_t)int_pack_size + // order
             (size_t)int_pack_size + // enforced_conserv
             (size_t)int_pack_size + // partial_coverage
             (size_t)int_pack_size;  // normalisation
    case (SOURCE_TO_TARGET_MAP):
      return (size_t)int_pack_size + // type
             (size_t)dbl_pack_size + // spread_distance
             (size_t)dbl_pack_size + // max_search_distance
             (size_t)int_pack_size;  // weight_type
    case (FIXED_VALUE):
      return (size_t)int_pack_size + // type
             (size_t)dbl_pack_size;  // value
    case (BERNSTEIN_BEZIER):
      return (size_t)int_pack_size;  // type
    case (USER_FILE):
      return (size_t)int_pack_size + // type
             yac_couple_config_get_string_pack_size(
               entry->user_file.filename, comm) +
             yac_couple_config_get_string_pack_size(
               entry->user_file.src_grid_name, comm) +
             yac_couple_config_get_string_pack_size(
               entry->user_file.tgt_grid_name, comm) +
             (size_t)int_pack_size;  // io_proc_step
    case (CHECK):
      return (size_t)int_pack_size + // type
             yac_couple_config_get_string_pack_size(
               entry->check.constructor_key, comm) + // constructor_key
             yac_couple_config_get_string_pack_size(
               entry->check.do_search_key, comm);    // do_search_key
    case (CREEP):
      return (size_t)int_pack_size + // type
             (size_t)int_pack_size;  // creep_distance
    case (USER_CALLBACK):
      return (size_t)int_pack_size + // type
             yac_couple_config_get_string_pack_size(
               entry->user_callback.func_compute_weights_key, comm);
               // func_compute_weights_key
  }
}

static size_t yac_couple_config_get_interp_stack_pack_size(
  struct yac_interp_stack_config interp_stack, MPI_Comm comm) {

  int size_pack_size;
  yac_mpi_call(MPI_Pack_size(1, MPI_INT, comm, &size_pack_size), comm);

  size_t config_pack_size = 0;

  for (size_t i = 0; i < interp_stack.size; ++i)
    config_pack_size +=
      yac_couple_config_get_interp_stack_entry_pack_size(
        interp_stack.config + i, comm);

  return (size_t)size_pack_size + config_pack_size;
}

static size_t yac_couple_config_get_transient_couple_pack_size(
  struct yac_couple_config_transient_couple * transient_couple, MPI_Comm comm) {

  int ints_pack_size;
  yac_mpi_call(MPI_Pack_size(9, MPI_INT, comm, &ints_pack_size), comm);

  return
    (size_t)ints_pack_size + // source.component_idx
                             // source.component_transient_idx
                             // source.lag
                             // target.component_idx
                             // target.component_transient_idx
                             // target.lag
                             // mapping_on_source
                             // coupling_period_operation
                             // enforce_write_weight_file
    yac_couple_config_get_string_pack_size(
      transient_couple->source.timestep, comm) +
    yac_couple_config_get_string_pack_size(
      transient_couple->target.timestep, comm) +
    yac_couple_config_get_interp_stack_pack_size(
      transient_couple->interp_stack, comm) +
    yac_couple_config_get_string_pack_size(
      transient_couple->coupling_period, comm) +
    yac_couple_config_get_string_pack_size(
      transient_couple->weight_file_name, comm);
}

static size_t yac_couple_config_get_transient_couples_pack_size(
  size_t num_transient_couples,
  struct yac_couple_config_transient_couple * transient_couples,
  MPI_Comm comm) {

  int num_transient_couples_pack_size;
  yac_mpi_call(
    MPI_Pack_size(1, MPI_INT, comm, &num_transient_couples_pack_size), comm);

  size_t transient_couples_pack_size = 0;
  for (size_t i = 0; i < num_transient_couples; ++i)
    transient_couples_pack_size +=
      yac_couple_config_get_transient_couple_pack_size(transient_couples + i, comm);

  return (size_t)num_transient_couples_pack_size +
         transient_couples_pack_size;
}

static size_t yac_couple_config_get_couple_pack_size(
  struct yac_couple_config_couple * couple, MPI_Comm comm) {

  int component_indices_pack_size;
  yac_mpi_call(
    MPI_Pack_size(2, MPI_INT, comm, &component_indices_pack_size), comm);

  return
    (size_t)component_indices_pack_size +
    yac_couple_config_get_transient_couples_pack_size(
      couple->num_transient_couples, couple->transient_couples, comm);
}

static size_t yac_couple_config_get_couples_pack_size(
  size_t num_couples, struct yac_couple_config_couple * couples,
  MPI_Comm comm) {

  int num_couples_pack_size;
  yac_mpi_call(
    MPI_Pack_size(1, MPI_INT, comm, &num_couples_pack_size), comm);

  size_t couples_pack_size = 0;
  for (size_t i = 0; i < num_couples; ++i)
    couples_pack_size +=
      yac_couple_config_get_couple_pack_size(couples + i, comm);

  return (size_t)num_couples_pack_size + couples_pack_size;
}

static size_t  yac_couple_config_get_pack_size(
  struct yac_couple_config * couple_config, MPI_Comm comm) {

  int int_pack_size;
  yac_mpi_call(MPI_Pack_size(1, MPI_INT, comm, &int_pack_size), comm);

  return (size_t)int_pack_size + // timestep_unit
         yac_couple_config_get_string_pack_size(
           couple_config->start_datetime, comm) +
         yac_couple_config_get_string_pack_size(
           couple_config->end_datetime, comm) +
         (size_t)int_pack_size + // calendar
         yac_couple_config_get_grids_pack_size(
           couple_config->num_grids, couple_config->grids, comm) +
         yac_couple_config_get_components_pack_size(
           couple_config->num_components, couple_config->components, comm) +
         yac_couple_config_get_transients_pack_size(
           couple_config->num_transients, couple_config->transients, comm) +
         yac_couple_config_get_couples_pack_size(
           couple_config->num_couples, couple_config->couples, comm) +
         (size_t)int_pack_size;  // redirect_stdout
}

static void yac_couple_config_pack_string(
  char const * string, void * buffer, int buffer_size, int * position,
  MPI_Comm comm) {

  size_t len = (string == NULL)?0:strlen(string);

  YAC_ASSERT(
    len <= INT_MAX, "ERROR(yac_couple_config_pack_string): string too long")

  int len_int = (int)len;

  yac_mpi_call(
    MPI_Pack(
      &len_int, 1, MPI_INT, buffer, buffer_size, position, comm), comm);

  if (len > 0)
    yac_mpi_call(
      MPI_Pack(
        string, len_int, MPI_CHAR, buffer, buffer_size, position, comm),
      comm);
}

static void yac_couple_config_pack_grid(
  struct yac_couple_config_grid * grid,
  void * buffer, int buffer_size, int * position, MPI_Comm comm) {

  yac_couple_config_pack_string(
    grid->name, buffer, buffer_size, position, comm);

  yac_mpi_call(
    MPI_Pack(
      &(grid->id), 1, MPI_INT, buffer, buffer_size, position, comm), comm);
}

static void yac_couple_config_pack_grids(
  size_t num_grids, struct yac_couple_config_grid * grids,
  void * buffer, int buffer_size, int * position, MPI_Comm comm) {

  YAC_ASSERT(
    num_grids <= INT_MAX,
    "ERROR(yac_couple_config_pack_grids): num_grids bigger than INT_MAX")

  int num_grids_int = (int)num_grids;
  yac_mpi_call(
    MPI_Pack(
      &num_grids_int, 1, MPI_INT, buffer, buffer_size, position, comm), comm);

  for (size_t i = 0; i < num_grids; ++i)
    yac_couple_config_pack_grid(
      grids + i, buffer, buffer_size, position, comm);
}

static void yac_couple_config_pack_component_transient(
  struct yac_couple_config_component_transient * component_transient,
  void * buffer, int buffer_size, int * position, MPI_Comm comm) {

  YAC_ASSERT(
    (component_transient->transient_idx >= 0) &&
    (component_transient->transient_idx <= INT_MAX),
    "ERROR(yac_couple_config_pack_component_transient):"
    "transient_idx is too big")

  YAC_ASSERT(
    (component_transient->grid_idx >= 0) &&
    (component_transient->grid_idx <= INT_MAX),
    "ERROR(yac_couple_config_pack_component_transient):"
    "grid_idx is too big")

  int indices[2] = {(int)(component_transient->transient_idx),
                    (int)(component_transient->grid_idx)};

  yac_mpi_call(
    MPI_Pack(
      indices, 2, MPI_INT, buffer, buffer_size, position, comm), comm);
}

static void yac_couple_config_pack_component_transients(
  size_t num_component_transients,
  struct yac_couple_config_component_transient * component_transients,
  void * buffer, int buffer_size, int * position, MPI_Comm comm) {

  YAC_ASSERT(
    num_component_transients <= INT_MAX,
    "ERROR(yac_couple_config_pack_component_transients): "
    "num_component_transients bigger than INT_MAX")

  int num_component_transients_int = (int)num_component_transients;
  yac_mpi_call(
    MPI_Pack(
      &num_component_transients_int, 1, MPI_INT,
      buffer, buffer_size, position, comm), comm);

  for (size_t i = 0; i < num_component_transients; ++i)
    yac_couple_config_pack_component_transient(
      component_transients + i, buffer, buffer_size, position, comm);
}

static void yac_couple_config_pack_component(
  struct yac_couple_config_component * component,
  void * buffer, int buffer_size, int * position, MPI_Comm comm) {

  yac_couple_config_pack_component_transients(
    component->num_component_transients, component->component_transients,
    buffer, buffer_size, position, comm);

  yac_couple_config_pack_string(
    component->name, buffer, buffer_size, position, comm);

  yac_mpi_call(
    MPI_Pack(
      &(component->id), 1, MPI_INT, buffer, buffer_size, position, comm), comm);
}

static void yac_couple_config_pack_components(
  size_t num_components, struct yac_couple_config_component * components,
  void * buffer, int buffer_size, int * position, MPI_Comm comm) {

  YAC_ASSERT(
    num_components <= INT_MAX,
    "ERROR(yac_couple_config_pack_components): "
    "num_components bigger than INT_MAX")

  int num_components_int = (int)num_components;
  yac_mpi_call(
    MPI_Pack(
      &num_components_int, 1, MPI_INT, buffer, buffer_size, position, comm), comm);

  for (size_t i = 0; i < num_components; ++i)
    yac_couple_config_pack_component(
      components + i, buffer, buffer_size, position, comm);
}

static void yac_couple_config_pack_transient(
  struct yac_couple_config_transient * transient,
  void * buffer, int buffer_size, int * position, MPI_Comm comm) {

  YAC_ASSERT(
    transient->collection_size <= INT_MAX,
    "ERROR(yac_couple_config_pack_components): "
    "collection_size bigger than INT_MAX")

  yac_couple_config_pack_string(
    transient->name, buffer, buffer_size, position, comm);

  int ints[2] = {transient->id, (int)(transient->collection_size)};
  yac_mpi_call(
    MPI_Pack(ints, 2, MPI_INT, buffer, buffer_size, position, comm), comm);
}

static void yac_couple_config_pack_transients(
  size_t num_transients, struct yac_couple_config_transient * transients,
  void * buffer, int buffer_size, int * position, MPI_Comm comm) {

  YAC_ASSERT(
    num_transients <= INT_MAX,
    "ERROR(yac_couple_config_pack_transients): "
    "num_transients bigger than INT_MAX")

  int num_transients_int = (int)num_transients;
  yac_mpi_call(
    MPI_Pack(
      &num_transients_int, 1, MPI_INT, buffer, buffer_size, position, comm), comm);

  for (size_t i = 0; i < num_transients; ++i)
    yac_couple_config_pack_transient(
      transients + i, buffer, buffer_size, position, comm);
}

static void yac_couple_config_pack_interp_stack_entry(
  union yac_interp_stack_config_entry * entry,
  void * buffer, int buffer_size, int * position, MPI_Comm comm) {

  int type = (int)(entry->general.type);
  yac_mpi_call(
    MPI_Pack(&type, 1, MPI_INT, buffer, buffer_size, position, comm), comm);

  check_interpolation_type(
    entry->general.type,
    "yac_couple_config_pack_interp_stack_entry");
  YAC_ASSERT(
    entry->general.type != UNDEFINED,
    "ERROR(yac_couple_config_pack_interp_stack_entry): "
    "invalid interpolation type")
  switch (entry->general.type) {
    default:
    case (AVERAGE): {
      int reduction_type = (int)(entry->average.reduction_type);
      yac_mpi_call(
        MPI_Pack(
          &reduction_type, 1, MPI_INT, buffer, buffer_size, position, comm),
        comm);
      yac_mpi_call(
        MPI_Pack(
          &(entry->average.partial_coverage), 1, MPI_INT,
          buffer, buffer_size, position, comm), comm);
      break;
    }
    case (RADIAL_BASIS_FUNCTION):
    case (N_NEAREST_NEIGHBOR): {
      YAC_ASSERT(
        entry->n_nearest_neighbor.config.n <= INT_MAX,
        "ERROR(yac_couple_config_pack_interp_stack_entry): "
        "n_nearest_neighbor.config.n bigger than INT_MAX")
      int type = (int)(entry->n_nearest_neighbor.config.type);
      yac_mpi_call(
        MPI_Pack(
          &type, 1, MPI_INT, buffer, buffer_size, position, comm), comm);
      int n = (int)(entry->n_nearest_neighbor.config.n);
      yac_mpi_call(
        MPI_Pack(
          &n, 1, MPI_INT, buffer, buffer_size, position, comm), comm);
      yac_mpi_call(
        MPI_Pack(
          &(entry->n_nearest_neighbor.config.data.rbf_scale), 1, MPI_DOUBLE,
          buffer, buffer_size, position, comm), comm);
      break;
    }
    case (CONSERVATIVE): {
      yac_mpi_call(
        MPI_Pack(
          &(entry->conservative.order), 1, MPI_INT,
          buffer, buffer_size, position, comm), comm);
      yac_mpi_call(
        MPI_Pack(
          &(entry->conservative.enforced_conserv), 1, MPI_INT,
          buffer, buffer_size, position, comm), comm);
      yac_mpi_call(
        MPI_Pack(
          &(entry->conservative.partial_coverage), 1, MPI_INT,
          buffer, buffer_size, position, comm), comm);
      int normalisation = (int)(entry->conservative.normalisation);
      yac_mpi_call(
        MPI_Pack(
          &normalisation, 1, MPI_INT, buffer, buffer_size, position, comm),
        comm);
      break;
    }
    case (SOURCE_TO_TARGET_MAP): {
      yac_mpi_call(
        MPI_Pack(
          &(entry->spmap.spread_distance), 1, MPI_DOUBLE,
          buffer, buffer_size, position, comm), comm);
      yac_mpi_call(
        MPI_Pack(
          &(entry->spmap.max_search_distance), 1, MPI_DOUBLE,
          buffer, buffer_size, position, comm), comm);
      int weight_type = (int)(entry->spmap.weight_type);
      yac_mpi_call(
        MPI_Pack(
          &weight_type, 1, MPI_INT, buffer, buffer_size, position, comm),
        comm);
      break;
    }
    case (FIXED_VALUE): {
      yac_mpi_call(
        MPI_Pack(
          &(entry->fixed.value), 1, MPI_DOUBLE,
          buffer, buffer_size, position, comm), comm);
      break;
    }
    case (BERNSTEIN_BEZIER):
      break;
    case (USER_FILE): {
      yac_couple_config_pack_string(
        entry->user_file.filename, buffer, buffer_size, position, comm);
      yac_couple_config_pack_string(
        entry->user_file.src_grid_name, buffer, buffer_size, position, comm);
      yac_couple_config_pack_string(
        entry->user_file.tgt_grid_name, buffer, buffer_size, position, comm);
      yac_mpi_call(
        MPI_Pack(
          &(entry->user_file.io_proc_step), 1, MPI_INT,
          buffer, buffer_size, position, comm), comm);
      break;
    }
    case (CHECK): {
      yac_couple_config_pack_string(
        entry->check.constructor_key, buffer, buffer_size, position, comm);
      yac_couple_config_pack_string(
        entry->check.do_search_key, buffer, buffer_size, position, comm);
      break;
    }
    case (CREEP): {
      yac_mpi_call(
        MPI_Pack(
          &(entry->creep.creep_distance), 1, MPI_INT,
          buffer, buffer_size, position, comm), comm);
      break;
    }
    case (USER_CALLBACK): {
      yac_couple_config_pack_string(
        entry->user_callback.func_compute_weights_key,
        buffer, buffer_size, position, comm);
      break;
    }
  }
}

static void yac_couple_config_pack_interp_stack(
  struct yac_interp_stack_config interp_stack,
  void * buffer, int buffer_size, int * position, MPI_Comm comm) {

  YAC_ASSERT(
    interp_stack.size <= INT_MAX,
    "ERROR(yac_couple_config_pack_interp_stack): "
    "interp_stack.size bigger than INT_MAX")

  int stack_size = (int)(interp_stack.size);
  yac_mpi_call(
    MPI_Pack(
      &stack_size, 1, MPI_INT,
      buffer, buffer_size, position, comm), comm);

  for (size_t i = 0; i < interp_stack.size; ++i)
    yac_couple_config_pack_interp_stack_entry(
      interp_stack.config + i, buffer, buffer_size, position, comm);
}

static void yac_couple_config_pack_transient_couple(
  struct yac_couple_config_transient_couple * transient_couple,
  void * buffer, int buffer_size, int * position, MPI_Comm comm) {

  YAC_ASSERT(
    (transient_couple->source.component_idx >= 0) &&
    (transient_couple->source.component_idx <= INT_MAX),
    "ERROR(yac_couple_config_pack_transient_couple): "
    "source.component_idx bigger than INT_MAX")
  YAC_ASSERT(
    (transient_couple->source.component_transient_idx >= 0) &&
    (transient_couple->source.component_transient_idx <= INT_MAX),
    "ERROR(yac_couple_config_pack_transient_couple): "
    "source.component_transient_idx bigger than INT_MAX")
  YAC_ASSERT(
    transient_couple->source.lag <= INT_MAX,
    "ERROR(yac_couple_config_pack_transient_couple): "
    "source.lag bigger than INT_MAX")
  YAC_ASSERT(
    (transient_couple->target.component_idx >= 0) &&
    (transient_couple->target.component_idx <= INT_MAX),
    "ERROR(yac_couple_config_pack_transient_couple): "
    "target.component_idx bigger than INT_MAX")
  YAC_ASSERT(
    (transient_couple->target.component_transient_idx >= 0) &&
    (transient_couple->target.component_transient_idx <= INT_MAX),
    "ERROR(yac_couple_config_pack_transient_couple): "
    "target.component_transient_idx bigger than INT_MAX")
  YAC_ASSERT(
    transient_couple->target.lag <= INT_MAX,
    "ERROR(yac_couple_config_pack_transient_couple): "
    "target.lag bigger than INT_MAX")
  YAC_ASSERT(
    transient_couple->mapping_on_source <= INT_MAX,
    "ERROR(yac_couple_config_pack_transient_couple): "
    "mapping_on_source bigger than INT_MAX")
  YAC_ASSERT(
    transient_couple->coupling_period_operation <= INT_MAX,
    "ERROR(yac_couple_config_pack_transient_couple): "
    "coupling_period_operation bigger than INT_MAX")
  YAC_ASSERT(
    transient_couple->enforce_write_weight_file <= INT_MAX,
    "ERROR(yac_couple_config_pack_transient_couple): "
    "enforce_write_weight_file bigger than INT_MAX")

  int ints[9] = {
    transient_couple->source.component_idx,
    transient_couple->source.component_transient_idx,
    transient_couple->source.lag,
    transient_couple->target.component_idx,
    transient_couple->target.component_transient_idx,
    transient_couple->target.lag,
    transient_couple->mapping_on_source,
    transient_couple->coupling_period_operation,
    transient_couple->enforce_write_weight_file};

  yac_mpi_call(
    MPI_Pack(ints, 9, MPI_INT, buffer, buffer_size, position, comm), comm);

  yac_couple_config_pack_string(
    transient_couple->source.timestep, buffer, buffer_size, position, comm);
  yac_couple_config_pack_string(
    transient_couple->target.timestep, buffer, buffer_size, position, comm);
  yac_couple_config_pack_string(
    transient_couple->coupling_period, buffer, buffer_size, position, comm);
  yac_couple_config_pack_string(
    transient_couple->weight_file_name, buffer, buffer_size, position, comm);

  yac_couple_config_pack_interp_stack(
    transient_couple->interp_stack, buffer, buffer_size, position, comm);
}

static void yac_couple_config_pack_transient_couples(
  size_t num_transient_couples, struct yac_couple_config_transient_couple * transient_couples,
  void * buffer, int buffer_size, int * position, MPI_Comm comm) {

  YAC_ASSERT(
    num_transient_couples <= INT_MAX,
    "ERROR(yac_couple_config_pack_transient_couples): "
    "num_transient_couples bigger than INT_MAX")

  int num_transient_couples_int = (int)num_transient_couples;
  yac_mpi_call(
    MPI_Pack(
      &num_transient_couples_int, 1, MPI_INT,
      buffer, buffer_size, position, comm), comm);

  for (size_t i = 0; i < num_transient_couples; ++i)
    yac_couple_config_pack_transient_couple(
      transient_couples + i, buffer, buffer_size, position, comm);
}

static void yac_couple_config_pack_couple(
  struct yac_couple_config_couple * couple,
  void * buffer, int buffer_size, int * position, MPI_Comm comm) {

  YAC_ASSERT(
    (couple->component_indices[0] <= INT_MAX) &&
    (couple->component_indices[1] <= INT_MAX),
    "ERROR(yac_couple_config_pack_couple): "
    "component_indices bigger than INT_MAX")

  int component_indices[2] = {(int)(couple->component_indices[0]),
                              (int)(couple->component_indices[1])};

  yac_mpi_call(
    MPI_Pack(
      component_indices, 2, MPI_INT, buffer, buffer_size, position, comm),
    comm);

  yac_couple_config_pack_transient_couples(
    couple->num_transient_couples, couple->transient_couples,
    buffer, buffer_size, position, comm);
}

static void yac_couple_config_pack_couples(
  size_t num_couples, struct yac_couple_config_couple * couples,
  void * buffer, int buffer_size, int * position, MPI_Comm comm) {

  YAC_ASSERT(
    num_couples <= INT_MAX,
    "ERROR(yac_couple_config_pack_couples): num_couples bigger than INT_MAX")

  int num_couples_int = (int)num_couples;
  yac_mpi_call(
    MPI_Pack(
      &num_couples_int, 1, MPI_INT, buffer, buffer_size, position, comm), comm);

  for (size_t i = 0; i < num_couples; ++i)
    yac_couple_config_pack_couple(
      couples + i, buffer, buffer_size, position, comm);
}

void yac_couple_config_pack(
  struct yac_couple_config * couple_config,
  int * buffer_size, void ** buffer, MPI_Comm comm) {

  size_t pack_size = yac_couple_config_get_pack_size(couple_config, comm);

  YAC_ASSERT(
    pack_size <= INT_MAX,
    "ERROR(yac_couple_config_pack): "
    "pack size of coupling configuration is too big")

  void * pack_buffer = xmalloc(pack_size);

  *buffer_size = (int)pack_size;
  *buffer = pack_buffer;

  int pack_size_int = (int)pack_size;
  int position = 0;

  int timestep_unit = (int)(couple_config->timestep_unit);
  yac_mpi_call(
    MPI_Pack(
      &timestep_unit, 1, MPI_INT, pack_buffer, pack_size_int, &position, comm),
    comm);
  yac_couple_config_pack_string(
    couple_config->start_datetime, pack_buffer, pack_size_int, &position,
    comm);
  yac_couple_config_pack_string(
    couple_config->end_datetime, pack_buffer, pack_size_int, &position,
    comm);
  int calendar = (int)couple_config->calendar;
  yac_mpi_call(
    MPI_Pack(
      &calendar, 1, MPI_INT, pack_buffer, pack_size_int, &position, comm),
    comm);

  yac_couple_config_pack_grids(
    couple_config->num_grids, couple_config->grids,
    pack_buffer, pack_size_int, &position, comm);
  yac_couple_config_pack_components(
    couple_config->num_components, couple_config->components,
    pack_buffer, pack_size_int, &position, comm);
  yac_couple_config_pack_transients(
    couple_config->num_transients, couple_config->transients,
    pack_buffer, pack_size_int, &position, comm);
  yac_couple_config_pack_couples(
    couple_config->num_couples, couple_config->couples,
    pack_buffer, pack_size_int, &position, comm);

  yac_mpi_call(
    MPI_Pack(
      &(couple_config->redirect_stdout), 1, MPI_INT,
      pack_buffer, pack_size_int, &position, comm), comm);
}

static char const * yac_couple_config_unpack_string(
  void * buffer, int buffer_size, int * position, MPI_Comm comm) {

  int string_len;
  yac_mpi_call(
    MPI_Unpack(
      buffer, buffer_size, position, &string_len, 1, MPI_INT, comm), comm);

  if (string_len <= 0) return NULL;

  char * string = xmalloc(((size_t)string_len + 1) * sizeof(*string));
  yac_mpi_call(
    MPI_Unpack(
      buffer, buffer_size, position, string, string_len, MPI_CHAR, comm), comm);
  string[string_len] = '\0';
  return string;
}

static void yac_couple_config_unpack_grid(
  void * buffer, int buffer_size, int * position,
  struct yac_couple_config_grid * grid, MPI_Comm comm) {

  grid->name =
    yac_couple_config_unpack_string(buffer, buffer_size, position, comm);
  yac_mpi_call(
    MPI_Unpack(
      buffer, buffer_size, position, &(grid->id), 1, MPI_INT, comm), comm);
}

static void yac_couple_config_unpack_grids(
  void * buffer, int buffer_size, int * position,
  size_t * num_grids, struct yac_couple_config_grid ** grids, MPI_Comm comm) {

  int num_grids_int;
  yac_mpi_call(
    MPI_Unpack(
      buffer, buffer_size, position, &num_grids_int, 1, MPI_INT, comm), comm);

  YAC_ASSERT(
    num_grids_int >= 0,
    "ERROR(yac_couple_config_unpack_grids): invalid number of grids")

  *num_grids = (size_t)num_grids_int;
  *grids = xmalloc(*num_grids * sizeof(**grids));

  for (size_t i = 0; i < *num_grids; ++i)
    yac_couple_config_unpack_grid(
      buffer, buffer_size, position, *grids + i, comm);
}

static void yac_couple_config_unpack_component_transient(
  void * buffer, int buffer_size, int * position,
  struct yac_couple_config_component_transient * component_transient, MPI_Comm comm) {

  int indices[2];

  yac_mpi_call(
    MPI_Unpack(
      buffer, buffer_size, position, indices, 2, MPI_INT, comm), comm);

  YAC_ASSERT(
    indices[0] >= 0,
    "ERROR(yac_couple_config_unpack_component_transient): "
    "invalid number of transient_idx")
  YAC_ASSERT(
    indices[1] >= 0,
    "ERROR(yac_couple_config_unpack_component_transient): "
    "invalid number of grid_idx")

  component_transient->transient_idx = (size_t)(indices[0]);
  component_transient->grid_idx = (size_t)(indices[1]);
}

static void yac_couple_config_unpack_component_transients(
  void * buffer, int buffer_size, int * position,
  size_t * num_component_transients,
  struct yac_couple_config_component_transient ** component_transients,
  MPI_Comm comm) {

  int num_component_transients_int;
  yac_mpi_call(
    MPI_Unpack(
      buffer, buffer_size, position, &num_component_transients_int, 1,
      MPI_INT, comm), comm);

  YAC_ASSERT(
    num_component_transients_int >= 0,
    "ERROR(yac_couple_config_unpack_component_transients): "
    "invalid number of component_transients")

  *num_component_transients = (size_t)num_component_transients_int;
  *component_transients =
    xmalloc(*num_component_transients * sizeof(**component_transients));

  for (size_t i = 0; i < *num_component_transients; ++i)
    yac_couple_config_unpack_component_transient(
      buffer, buffer_size, position, *component_transients + i, comm);
}

static void yac_couple_config_unpack_component(
  void * buffer, int buffer_size, int * position,
  struct yac_couple_config_component * component, MPI_Comm comm) {

  yac_couple_config_unpack_component_transients(
    buffer, buffer_size, position, &(component->num_component_transients),
    &(component->component_transients), comm);
  component->name =
    yac_couple_config_unpack_string(buffer, buffer_size, position, comm);
  yac_mpi_call(
    MPI_Unpack(
      buffer, buffer_size, position, &(component->id), 1, MPI_INT, comm),
    comm);
}

static void yac_couple_config_unpack_components(
  void * buffer, int buffer_size, int * position,
  size_t * num_components, struct yac_couple_config_component ** components,
  MPI_Comm comm) {

  int num_components_int;
  yac_mpi_call(
    MPI_Unpack(
      buffer, buffer_size, position, &num_components_int, 1, MPI_INT, comm),
    comm);

  YAC_ASSERT(
    num_components_int >= 0,
    "ERROR(yac_couple_config_unpack_components): "
    "invalid number of components")

  *num_components = (size_t)num_components_int;
  *components = xmalloc(*num_components * sizeof(**components));

  for (size_t i = 0; i < *num_components; ++i)
    yac_couple_config_unpack_component(
      buffer, buffer_size, position, *components + i, comm);
}

static void yac_couple_config_unpack_transient(
  void * buffer, int buffer_size, int * position,
  struct yac_couple_config_transient * transient, MPI_Comm comm) {

  transient->name =
    yac_couple_config_unpack_string(buffer, buffer_size, position, comm);

  int ints[2];
  yac_mpi_call(
    MPI_Unpack(
      buffer, buffer_size, position, ints, 2, MPI_INT, comm), comm);

  YAC_ASSERT(
    ints[1] >= 0,
    "ERROR(yac_couple_config_unpack_transient): "
    "invalid collection size")

  transient->id = ints[0];
  transient->collection_size = ints[1];
}

static void yac_couple_config_unpack_transients(
  void * buffer, int buffer_size, int * position,
  size_t * num_transients, struct yac_couple_config_transient ** transients,
  MPI_Comm comm) {

  int num_transients_int;
  yac_mpi_call(
    MPI_Unpack(
      buffer, buffer_size, position, &num_transients_int, 1, MPI_INT, comm),
    comm);

  YAC_ASSERT(
    num_transients_int >= 0,
    "ERROR(yac_couple_config_unpack_transients): "
    "invalid number of transients")

  *num_transients = (size_t)num_transients_int;
  *transients = xmalloc(*num_transients * sizeof(**transients));

  for (size_t i = 0; i < *num_transients; ++i)
    yac_couple_config_unpack_transient(
      buffer, buffer_size, position, *transients + i, comm);
}

static void yac_couple_config_unpack_n_string(
  void * buffer, int buffer_size, int * position,
  char * string, int max_string_len, MPI_Comm comm) {

  int string_len;
  yac_mpi_call(
    MPI_Unpack(
      buffer, buffer_size, position, &string_len, 1, MPI_INT, comm), comm);

  YAC_ASSERT(
    string_len >= 0,
    "ERROR(yac_couple_config_unpack_n_string): invalid string length")

  YAC_ASSERT(
    string_len < max_string_len,
    "ERROR(yac_couple_config_unpack_n_string): string length to long")

  if (string_len > 0)
    yac_mpi_call(
      MPI_Unpack(
        buffer, buffer_size, position, string, string_len, MPI_CHAR, comm),
      comm);
  string[string_len] = '\0';
}

static void yac_couple_config_unpack_interp_stack_entry(
  void * buffer, int buffer_size, int * position,
  union yac_interp_stack_config_entry * entry, MPI_Comm comm) {

  int type;
  yac_mpi_call(
    MPI_Unpack(
      buffer, buffer_size, position, &type, 1, MPI_INT, comm), comm);

  entry->general.type = (enum yac_interpolation_list)type;

  check_interpolation_type(
    entry->general.type,
    "yac_couple_config_unpack_interp_stack_entry");
  YAC_ASSERT(
    entry->general.type != UNDEFINED,
    "ERROR(yac_couple_config_unpack_interp_stack_entry): "
    "invalid interpolation type")
  switch (type) {
    default:
    case (AVERAGE): {
      int reduction_type;
      yac_mpi_call(
        MPI_Unpack(
          buffer, buffer_size, position, &reduction_type, 1, MPI_INT, comm),
        comm);
      entry->average.reduction_type =
        (enum yac_interp_avg_weight_type)reduction_type;
      yac_mpi_call(
        MPI_Unpack(
          buffer, buffer_size, position, &(entry->average.partial_coverage),
          1, MPI_INT, comm), comm);
      break;
    }
    case (RADIAL_BASIS_FUNCTION):
    case (N_NEAREST_NEIGHBOR): {
      int type;
      yac_mpi_call(
        MPI_Unpack(
          buffer, buffer_size, position, &type, 1, MPI_INT, comm), comm);
      entry->n_nearest_neighbor.config.type =
        (enum yac_interp_nnn_weight_type)type;
      int n;
      yac_mpi_call(
        MPI_Unpack(
          buffer, buffer_size, position, &n, 1, MPI_INT, comm), comm);
      YAC_ASSERT(
        n >= 0,
        "ERROR(yac_couple_config_unpack_interp_stack_entry): "
        "invalid n_nearest_neighbor.config.n")
      entry->n_nearest_neighbor.config.n = (size_t)n;
      yac_mpi_call(
        MPI_Unpack(
          buffer, buffer_size, position,
          &(entry->n_nearest_neighbor.config.data.rbf_scale),
          1, MPI_DOUBLE, comm), comm);
      break;
    }
    case (CONSERVATIVE): {
      yac_mpi_call(
        MPI_Unpack(
          buffer, buffer_size, position,
          &(entry->conservative.order), 1, MPI_INT, comm), comm);
      yac_mpi_call(
        MPI_Unpack(
          buffer, buffer_size, position,
          &(entry->conservative.enforced_conserv), 1, MPI_INT, comm), comm);
      yac_mpi_call(
        MPI_Unpack(
          buffer, buffer_size, position,
          &(entry->conservative.partial_coverage), 1, MPI_INT, comm), comm);
      int normalisation;
      yac_mpi_call(
        MPI_Unpack(
          buffer, buffer_size, position, &normalisation, 1, MPI_INT, comm),
        comm);
      entry->conservative.normalisation =
        (enum yac_interp_method_conserv_normalisation)normalisation;
      break;
    }
    case (SOURCE_TO_TARGET_MAP): {
      yac_mpi_call(
        MPI_Unpack(
          buffer, buffer_size, position,
          &(entry->spmap.spread_distance), 1, MPI_DOUBLE, comm), comm);
      yac_mpi_call(
        MPI_Unpack(
          buffer, buffer_size, position,
          &(entry->spmap.max_search_distance), 1, MPI_DOUBLE, comm), comm);
      int weight_type;
      yac_mpi_call(
        MPI_Unpack(
          buffer, buffer_size, position, &weight_type, 1, MPI_INT, comm),
        comm);
      entry->spmap.weight_type =
        (enum yac_interp_spmap_weight_type)weight_type;
      break;
    }
    case (FIXED_VALUE): {
      yac_mpi_call(
        MPI_Unpack(
          buffer, buffer_size, position,
          &(entry->fixed.value), 1, MPI_DOUBLE, comm), comm);
      break;
    }
    case (BERNSTEIN_BEZIER):
      break;
    case (USER_FILE): {
      yac_couple_config_unpack_n_string(
        buffer, buffer_size, position,
        entry->user_file.filename, MAX_FILE_NAME_LENGTH, comm);
      yac_couple_config_unpack_n_string(
        buffer, buffer_size, position,
        entry->user_file.src_grid_name, MAX_FILE_NAME_LENGTH, comm);
      yac_couple_config_unpack_n_string(
        buffer, buffer_size, position,
        entry->user_file.tgt_grid_name, MAX_FILE_NAME_LENGTH, comm);
      yac_mpi_call(
        MPI_Unpack(
          buffer, buffer_size, position,
          &(entry->user_file.io_proc_step), 1, MPI_INT, comm), comm);
      break;
    }
    case (CHECK): {
      yac_couple_config_unpack_n_string(
        buffer, buffer_size, position,
        entry->check.constructor_key, MAX_ROUTINE_NAME_LENGTH, comm);
      yac_couple_config_unpack_n_string(
        buffer, buffer_size, position,
        entry->check.do_search_key, MAX_ROUTINE_NAME_LENGTH, comm);
      break;
    }
    case (CREEP): {
      yac_mpi_call(
        MPI_Unpack(
          buffer, buffer_size, position,
          &(entry->creep.creep_distance), 1, MPI_INT, comm), comm);
      break;
    }
    case (USER_CALLBACK): {
      yac_couple_config_unpack_n_string(
        buffer, buffer_size, position,
        entry->user_callback.func_compute_weights_key,
        MAX_ROUTINE_NAME_LENGTH, comm);
      break;
    }
  }
}

static struct yac_interp_stack_config yac_couple_config_unpack_interp_stack(
  void * buffer, int buffer_size, int * position, MPI_Comm comm) {

  int stack_size;
  yac_mpi_call(
    MPI_Unpack(
      buffer, buffer_size, position, &stack_size, 1, MPI_INT, comm), comm);

  YAC_ASSERT(
    stack_size >= 0,
    "ERROR(yac_couple_config_unpack_interp_stack): "
    "invalid stack size")

  struct yac_interp_stack_config interp_stack;

  interp_stack.size = (size_t)stack_size;
  interp_stack.config =
    xmalloc((size_t)stack_size * sizeof(*interp_stack.config));

  for (int i = 0; i < stack_size; ++i)
    yac_couple_config_unpack_interp_stack_entry(
      buffer, buffer_size, position, interp_stack.config + (size_t)i, comm);

  return interp_stack;
}

static void yac_couple_config_unpack_transient_couple(
  void * buffer, int buffer_size, int * position,
  struct yac_couple_config_transient_couple * transient_couple,
  MPI_Comm comm) {

  int ints[9];
  yac_mpi_call(
    MPI_Unpack(
      buffer, buffer_size, position, ints, 9, MPI_INT, comm), comm);

  YAC_ASSERT(
    ints[0] >= 0,
    "ERROR(yac_couple_config_unpack_transient_couple): "
    "invalid source.component_idx")
  YAC_ASSERT(
    ints[1] >= 0,
    "ERROR(yac_couple_config_unpack_transient_couple): "
    "invalid source.component_transient_idx")
  YAC_ASSERT(
    ints[2] >= 0,
    "ERROR(yac_couple_config_unpack_transient_couple): "
    "invalid source.lag")
  YAC_ASSERT(
    ints[3] >= 0,
    "ERROR(yac_couple_config_unpack_transient_couple): "
    "target.component_idx bigger than INT_MAX")
  YAC_ASSERT(
    ints[4] >= 0,
    "ERROR(yac_couple_config_unpack_transient_couple): "
    "invalid target.component_transient_idx")
  YAC_ASSERT(
    ints[5] >= 0,
    "ERROR(yac_couple_config_unpack_transient_couple): "
    "invalid target.lag")
  YAC_ASSERT(
    ints[6] >= 0,
    "ERROR(yac_couple_config_unpack_transient_couple): "
    "invalid mapping_on_source")
  YAC_ASSERT(
    ints[7] >= 0,
    "ERROR(yac_couple_config_unpack_transient_couple): "
    "invalid coupling_period_operation")
  YAC_ASSERT(
    ints[8] >= 0,
    "ERROR(yac_couple_config_unpack_transient_couple): "
    "invalid enforce_write_weight_file")

  transient_couple->source.component_idx = (size_t)(ints[0]);
  transient_couple->source.component_transient_idx = (size_t)(ints[1]);
  transient_couple->source.lag = ints[2];
  transient_couple->target.component_idx = (size_t)(ints[3]);
  transient_couple->target.component_transient_idx = (size_t)(ints[4]);
  transient_couple->target.lag = ints[5];
  transient_couple->mapping_on_source = ints[6];
  transient_couple->coupling_period_operation =
    (enum yac_reduction_type)(ints[7]);
  transient_couple->enforce_write_weight_file = ints[8];

  transient_couple->source.timestep =
    yac_couple_config_unpack_string(buffer, buffer_size, position, comm);
  transient_couple->target.timestep =
    yac_couple_config_unpack_string(buffer, buffer_size, position, comm);
  transient_couple->coupling_period =
    yac_couple_config_unpack_string(buffer, buffer_size, position, comm);
  transient_couple->weight_file_name =
    yac_couple_config_unpack_string(buffer, buffer_size, position, comm);

  transient_couple->interp_stack =
    yac_couple_config_unpack_interp_stack(buffer, buffer_size, position, comm);
}

static void yac_couple_config_unpack_transient_couples(
  void * buffer, int buffer_size, int * position,
  size_t * num_transient_couples,
  struct yac_couple_config_transient_couple ** transient_couples,
  MPI_Comm comm) {

  int num_transient_couples_int;
  yac_mpi_call(
    MPI_Unpack(
      buffer, buffer_size, position, &num_transient_couples_int, 1,
      MPI_INT, comm), comm);

  YAC_ASSERT(
    num_transient_couples_int >= 0,
    "ERROR(yac_couple_config_unpack_transient_couples): "
    "invalid number of transient_couples")

  *num_transient_couples = (size_t)num_transient_couples_int;
  *transient_couples =
    xmalloc(*num_transient_couples * sizeof(**transient_couples));

  for (size_t i = 0; i < *num_transient_couples; ++i)
    yac_couple_config_unpack_transient_couple(
      buffer, buffer_size, position, *transient_couples + i, comm);
}

static void yac_couple_config_unpack_couple(
  void * buffer, int buffer_size, int * position,
  struct yac_couple_config_couple * couple, MPI_Comm comm) {

  int component_indices[2];
  yac_mpi_call(
    MPI_Unpack(
      buffer, buffer_size, position, component_indices, 2, MPI_INT, comm),
    comm);

  YAC_ASSERT(
    (component_indices[0] >= 0) && (component_indices[1] >= 0),
    "ERROR(yac_couple_config_unpack_couple): invalid component indices")

  couple->component_indices[0] = (size_t)(component_indices[0]);
  couple->component_indices[1] = (size_t)(component_indices[1]);

  yac_couple_config_unpack_transient_couples(
    buffer, buffer_size, position,
    &(couple->num_transient_couples), &(couple->transient_couples), comm);
}

static void yac_couple_config_unpack_couples(
  void * buffer, int buffer_size, int * position,
  size_t * num_couples, struct yac_couple_config_couple ** couples,
  MPI_Comm comm) {

  int num_couples_int;
  yac_mpi_call(
    MPI_Unpack(
      buffer, buffer_size, position, &num_couples_int, 1, MPI_INT, comm),
    comm);

  YAC_ASSERT(
    num_couples_int >= 0,
    "ERROR(yac_couple_config_unpack_couples): invalid number of couples")

  *num_couples = (size_t)num_couples_int;
  *couples = xmalloc(*num_couples * sizeof(**couples));

  for (size_t i = 0; i < *num_couples; ++i)
    yac_couple_config_unpack_couple(
      buffer, buffer_size, position, *couples + i, comm);
}

struct yac_couple_config * yac_couple_config_unpack(
  void * buffer, int buffer_size, MPI_Comm comm) {

  struct yac_couple_config * couple_config =
    xmalloc(1 * sizeof(*couple_config));

  int position = 0;

  int timestep_unit;
  yac_mpi_call(
    MPI_Unpack(
      buffer, buffer_size, &position, &timestep_unit, 1, MPI_INT, comm),
    comm);
  couple_config->timestep_unit = (enum yac_time_unit_type)timestep_unit;
  couple_config->start_datetime =
    yac_couple_config_unpack_string(buffer, buffer_size, &position, comm);
  couple_config->end_datetime =
    yac_couple_config_unpack_string(buffer, buffer_size, &position, comm);
  int calendar;
  yac_mpi_call(
    MPI_Unpack(
      buffer, buffer_size, &position, &calendar, 1, MPI_INT, comm),
    comm);
  couple_config->calendar = (calendarType)calendar;

  yac_couple_config_unpack_grids(
    buffer, buffer_size, &position,
    &(couple_config->num_grids), &(couple_config->grids), comm);
  yac_couple_config_unpack_components(
    buffer, buffer_size, &position,
    &(couple_config->num_components), &(couple_config->components), comm);
  yac_couple_config_unpack_transients(
    buffer, buffer_size, &position,
    &(couple_config->num_transients), &(couple_config->transients), comm);
  yac_couple_config_unpack_couples(
    buffer, buffer_size, &position,
    &(couple_config->num_couples), &(couple_config->couples), comm);

  yac_mpi_call(
    MPI_Unpack(
      buffer, buffer_size, &position,
      &(couple_config->redirect_stdout), 1, MPI_INT, comm), comm);

  return couple_config;
}


struct yac_couple_config * yac_couple_config_bcast(
  struct yac_couple_config * couple_config, int root, MPI_Comm comm) {

  int rank;
  MPI_Comm_rank(comm, &rank);

  void * buffer;
  int buffer_size;
  if (rank == root)
    yac_couple_config_pack(couple_config, &buffer_size, &buffer, comm);

  yac_mpi_call(MPI_Bcast(&buffer_size, 1, MPI_INT, root, comm), comm);

  if (rank != root) buffer = xmalloc((size_t)buffer_size);

  yac_mpi_call(MPI_Bcast(buffer, buffer_size, MPI_PACKED, root, comm), comm);

  if (rank == root)
    couple_config = yac_couple_config_copy(couple_config);
  else
    couple_config = yac_couple_config_unpack(buffer, buffer_size, comm);
  free(buffer);

  return couple_config;
}

struct yac_interp_stack_config yac_interp_stack_config_new() {

  return (struct yac_interp_stack_config){.config = NULL, .size = 0};
}
void yac_interp_stack_config_delete(
  struct yac_interp_stack_config interp_stack_config) {
  free(interp_stack_config.config);
}

static union yac_interp_stack_config_entry *
  yac_interp_stack_config_add_entry(
    struct yac_interp_stack_config * interp_stack_config) {

  interp_stack_config->size++;
  interp_stack_config->config =
    xrealloc(
      interp_stack_config->config,
      interp_stack_config->size * sizeof(*(interp_stack_config->config)));

  return interp_stack_config->config + (interp_stack_config->size - 1);
}

void yac_interp_stack_config_add_average(
  struct yac_interp_stack_config * interp_stack_config,
  enum yac_interp_avg_weight_type reduction_type, int partial_coverage) {

  union yac_interp_stack_config_entry * entry =
    yac_interp_stack_config_add_entry(interp_stack_config);

  entry->average.type = AVERAGE;
  entry->average.reduction_type = reduction_type;
  entry->average.partial_coverage = partial_coverage;
}

void yac_interp_stack_config_add_nnn(
  struct yac_interp_stack_config * interp_stack_config,
  enum yac_interp_nnn_weight_type type, size_t n, double scale) {

  union yac_interp_stack_config_entry * entry =
    yac_interp_stack_config_add_entry(interp_stack_config);

  if (type == NNN_RBF) {
    entry->radial_basis_function.type = RADIAL_BASIS_FUNCTION;
    entry->radial_basis_function.config =
      (struct yac_nnn_config){
        .type = NNN_RBF, .n = n, .data.rbf_scale = scale};
  } else {
    entry->n_nearest_neighbor.type = N_NEAREST_NEIGHBOR;
    entry->n_nearest_neighbor.config =
      (struct yac_nnn_config){
        .type = type, .n = n, .data.rbf_scale = scale};
  }
}

void yac_interp_stack_config_add_conservative(
  struct yac_interp_stack_config * interp_stack_config,
  int order, int enforced_conserv, int partial_coverage,
  enum yac_interp_method_conserv_normalisation normalisation) {

  union yac_interp_stack_config_entry * entry =
    yac_interp_stack_config_add_entry(interp_stack_config);

  entry->conservative.type = CONSERVATIVE;
  entry->conservative.order = order;
  entry->conservative.enforced_conserv = enforced_conserv;
  entry->conservative.partial_coverage = partial_coverage;
  entry->conservative.normalisation = normalisation;
}

void yac_interp_stack_config_add_spmap(
  struct yac_interp_stack_config * interp_stack_config,
  double spread_distance, double max_search_distance,
  enum yac_interp_spmap_weight_type weight_type) {

  union yac_interp_stack_config_entry * entry =
    yac_interp_stack_config_add_entry(interp_stack_config);

  entry->spmap.type = SOURCE_TO_TARGET_MAP;
  entry->spmap.spread_distance = spread_distance;
  entry->spmap.max_search_distance = max_search_distance;
  entry->spmap.weight_type = weight_type;
}

void yac_interp_stack_config_add_hcsbb(
  struct yac_interp_stack_config * interp_stack_config) {

  union yac_interp_stack_config_entry * entry =
    yac_interp_stack_config_add_entry(interp_stack_config);

  entry->hcsbb.type = BERNSTEIN_BEZIER;
}

static void check_string(
  char * string, char const * file, int line, char const * routine,
  char const * variable) {

  YAC_ASSERT_F(
    string != NULL, "ERROR(%s:%d:%s): %s is NULL",
    file, line, routine, variable)
  YAC_ASSERT_F(
    strlen(string) < MAX_FILE_NAME_LENGTH, "ERROR(%s:%d:%s): %s is too long",
    file, line, routine, variable)
}

void yac_interp_stack_config_add_user_file(
  struct yac_interp_stack_config * interp_stack_config,
  char * filename, char * src_grid_name, char * tgt_grid_name,
  int io_proc_step) {

  check_string(
    filename, __FILE__, __LINE__,
    "yac_interp_stack_config_add_user_file", "filename");
  check_string(
    src_grid_name, __FILE__, __LINE__,
    "yac_interp_stack_config_add_user_file", "src_grid_name");
  check_string(
    tgt_grid_name, __FILE__, __LINE__,
    "yac_interp_stack_config_add_user_file", "tgt_grid_name");

  union yac_interp_stack_config_entry * entry =
    yac_interp_stack_config_add_entry(interp_stack_config);

  entry->user_file.type = USER_FILE;
  strcpy(entry->user_file.filename, filename);
  strcpy(entry->user_file.src_grid_name, src_grid_name);
  strcpy(entry->user_file.tgt_grid_name, tgt_grid_name);
  entry->user_file.io_proc_step = io_proc_step;
}

void yac_interp_stack_config_add_fixed(
  struct yac_interp_stack_config * interp_stack_config, double value) {

  union yac_interp_stack_config_entry * entry =
    yac_interp_stack_config_add_entry(interp_stack_config);

  entry->fixed.type = FIXED_VALUE;
  entry->fixed.value = value;
}

void yac_interp_stack_config_add_check(
  struct yac_interp_stack_config * interp_stack_config,
  char * constructor_key, char * do_search_key) {

  check_string(
    constructor_key, __FILE__, __LINE__,
    "yac_interp_stack_config_add_check", "constructor_key");
  check_string(
    do_search_key, __FILE__, __LINE__,
    "yac_interp_stack_config_add_check", "do_search_key");

  union yac_interp_stack_config_entry * entry =
    yac_interp_stack_config_add_entry(interp_stack_config);

  entry->check.type = CHECK;
  strcpy(entry->check.constructor_key, constructor_key);
  strcpy(entry->check.do_search_key, do_search_key);
}

void yac_interp_stack_config_add_creep(
  struct yac_interp_stack_config * interp_stack_config, int creep_distance) {

  union yac_interp_stack_config_entry * entry =
    yac_interp_stack_config_add_entry(interp_stack_config);

  entry->creep.type = CREEP;
  entry->creep.creep_distance = creep_distance;
}

void yac_interp_stack_config_add_user_callback(
  struct yac_interp_stack_config * interp_stack_config,
  char * func_compute_weights_key) {

  check_string(
    func_compute_weights_key, __FILE__, __LINE__,
    "yac_interp_stack_config_add_user_callback",
    "func_compute_weights_key");

  union yac_interp_stack_config_entry * entry =
    yac_interp_stack_config_add_entry(interp_stack_config);

  entry->user_callback.type = USER_CALLBACK;
  strcpy(
    entry->user_callback.func_compute_weights_key, func_compute_weights_key);
}
