/**
 * @file test_couple_config.c
 *
 * @copyright Copyright  (C)  2021 DKRZ, MPI-M
 *
 * @author Teresa Holfeld <teresa.holfeld@mpimet.mpg.de>
 *         Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "tests.h"
#include "utils.h"
#include "config_xml.h"
#include "yac_mpi.h"

static void check_couple_config(struct yac_couple_config * couple_config);
static struct yac_couple_config * generate_couple_config_via_UI();

int main(void) {

  yac_mpi_call(MPI_Init(NULL, NULL), MPI_COMM_WORLD);

  int size, rank;
  MPI_Comm_rank ( MPI_COMM_WORLD, &rank );
  MPI_Comm_size ( MPI_COMM_WORLD, &size );

  if (size != 2) {
    fputs("ERROR wrong number of processes (has to be 2)", stderr);
    exit(EXIT_FAILURE);
  }

  check_couple_config(
    yac_couple_config_xml_new(
      "couple_config_test.xml", "coupling.xsd"));

  struct yac_couple_config * couple_config_UI =
    generate_couple_config_via_UI();

  { // check packing and unpacking of the coupling configuration
    void * buffer;
    int buffer_size;

    yac_couple_config_pack(
      couple_config_UI, &buffer_size, &buffer, MPI_COMM_WORLD);
    struct yac_couple_config * couple_config_unpacked =
      yac_couple_config_unpack(buffer, buffer_size, MPI_COMM_WORLD);

    check_couple_config(couple_config_unpacked);

    free(buffer);
  }

  check_couple_config(yac_couple_config_copy(couple_config_UI));

  check_couple_config(
    yac_couple_config_bcast(
      (rank == 0)?yac_couple_config_copy(couple_config_UI):NULL,
      0, MPI_COMM_WORLD));

  check_couple_config(couple_config_UI);

  yac_mpi_call(MPI_Finalize(), MPI_COMM_WORLD);

  return TEST_EXIT_CODE;
}

static void check_couple_config(struct yac_couple_config * couple_config) {

  if (yac_couple_config_redirect_stdout(couple_config))
    PUT_ERR("ERROR in yac_couple_config_redirect_stdout\n");

  if (yac_couple_config_get_num_components(couple_config) != 4)
    PUT_ERR("ERROR in yac_couple_config_get_num_components\n");

  if (!yac_couple_config_component_name_is_valid(couple_config, "ICON-O"))
    PUT_ERR("ERROR in yac_couple_config_component_name_is_valid\n");
  if (!yac_couple_config_component_name_is_valid(couple_config, "ICON-A"))
    PUT_ERR("ERROR in yac_couple_config_component_name_is_valid\n");
  if (!yac_couple_config_component_name_is_valid(couple_config, "DUMMY"))
    PUT_ERR("ERROR in yac_couple_config_component_name_is_valid\n");
  if (!yac_couple_config_component_name_is_valid(couple_config, "DUMMY_2"))
    PUT_ERR("ERROR in yac_couple_config_component_name_is_valid\n");
  if (yac_couple_config_component_name_is_valid(couple_config, "INVALID"))
    PUT_ERR("ERROR in yac_couple_config_component_name_is_valid\n");

  if (yac_couple_config_get_component_id(couple_config, 0) != 1)
    PUT_ERR("ERROR in yac_couple_config_get_component_id\n");
  if (yac_couple_config_get_component_id(couple_config, 1) != 2)
    PUT_ERR("ERROR in yac_couple_config_get_component_id\n");
  if (yac_couple_config_get_component_id(couple_config, 2) != 5)
    PUT_ERR("ERROR in yac_couple_config_get_component_id\n");
  if (yac_couple_config_get_component_id(couple_config, 3) != 3)
    PUT_ERR("ERROR in yac_couple_config_get_component_id\n");

  if (strcmp(yac_couple_config_get_component_name(couple_config, 0), "ICON-O"))
    PUT_ERR("ERROR in yac_couple_config_get_component_name\n");
  if (strcmp(yac_couple_config_get_component_name(couple_config, 1), "ICON-A"))
    PUT_ERR("ERROR in yac_couple_config_get_component_name\n");
  if (strcmp(yac_couple_config_get_component_name(couple_config, 2), "DUMMY"))
    PUT_ERR("ERROR in yac_couple_config_get_component_name\n");
  if (strcmp(yac_couple_config_get_component_name(couple_config, 3), "DUMMY_2"))
    PUT_ERR("ERROR in yac_couple_config_get_component_name\n");

  if (strcmp(yac_couple_config_get_component_name_by_id(couple_config, 1), "ICON-O"))
    PUT_ERR("ERROR in yac_couple_config_get_component_name_by_id\n");
  if (strcmp(yac_couple_config_get_component_name_by_id(couple_config, 2), "ICON-A"))
    PUT_ERR("ERROR in yac_couple_config_get_component_name_by_id\n");
  if (strcmp(yac_couple_config_get_component_name_by_id(couple_config, 5), "DUMMY"))
    PUT_ERR("ERROR in yac_couple_config_get_component_name_by_id\n");
  if (strcmp(yac_couple_config_get_component_name_by_id(couple_config, 3), "DUMMY_2"))
    PUT_ERR("ERROR in yac_couple_config_get_component_name_by_id\n");

  if (yac_couple_config_get_component_id_by_name(couple_config, "ICON-O") != 1)
    PUT_ERR("ERROR in yac_couple_config_get_component_id_by_name\n");
  if (yac_couple_config_get_component_id_by_name(couple_config, "ICON-A") != 2)
    PUT_ERR("ERROR in yac_couple_config_get_component_id_by_name\n");
  if (yac_couple_config_get_component_id_by_name(couple_config, "DUMMY") != 5)
    PUT_ERR("ERROR in yac_couple_config_get_component_id_by_name\n");
  if (yac_couple_config_get_component_id_by_name(couple_config, "DUMMY_2") != 3)
    PUT_ERR("ERROR in yac_couple_config_get_component_id_by_name\n");

  if (yac_couple_config_get_grid_id_by_name(couple_config, "grid1") != 1)
    PUT_ERR("ERROR in yac_couple_config_get_grid_id_by_name\n");
  if (yac_couple_config_get_grid_id_by_name(couple_config, "grid2") != 2)
    PUT_ERR("ERROR in yac_couple_config_get_grid_id_by_name\n");
  if (yac_couple_config_get_grid_id_by_name(couple_config, "grid3") != 3)
    PUT_ERR("ERROR in yac_couple_config_get_grid_id_by_name\n");
  if (yac_couple_config_get_grid_id_by_name(couple_config, "grid4") != 4)
    PUT_ERR("ERROR in yac_couple_config_get_grid_id_by_name\n");

  if (!yac_couple_config_contains_grid_name(couple_config, "grid1"))
    PUT_ERR("ERROR in yac_couple_config_contains_grid_name\n");
  if (!yac_couple_config_contains_grid_name(couple_config, "grid2"))
    PUT_ERR("ERROR in yac_couple_config_contains_grid_name\n");
  if (!yac_couple_config_contains_grid_name(couple_config, "grid3"))
    PUT_ERR("ERROR in yac_couple_config_contains_grid_name\n");
  if (!yac_couple_config_contains_grid_name(couple_config, "grid4"))
    PUT_ERR("ERROR in yac_couple_config_contains_grid_name\n");
  if (yac_couple_config_contains_grid_name(couple_config, "grid5"))
    PUT_ERR("ERROR in yac_couple_config_contains_grid_name\n");

  if (yac_couple_config_get_transient_id_by_name(
        couple_config, "sea_surface_temperature") != 1)
    PUT_ERR("ERROR in yac_couple_config_get_transient_id_by_name\n");
  if (yac_couple_config_get_transient_id_by_name(
        couple_config, "wind_speed") != 2)
    PUT_ERR("ERROR in yac_couple_config_get_transient_id_by_name\n");
  if (yac_couple_config_get_transient_id_by_name(
        couple_config, "water_flux_into_sea_water") != 3)
    PUT_ERR("ERROR in yac_couple_config_get_transient_id_by_name\n");
  if (yac_couple_config_get_transient_id_by_name(
        couple_config, "grid_eastward_wind") != 4)
    PUT_ERR("ERROR in yac_couple_config_get_transient_id_by_name\n");
  if (yac_couple_config_get_transient_id_by_name(
        couple_config, "grid_northward_wind") != 5)
    PUT_ERR("ERROR in yac_couple_config_get_transient_id_by_name\n");
  if (yac_couple_config_get_transient_id_by_name(
        couple_config, "dummy_field_1") != 6)
    PUT_ERR("ERROR in yac_couple_config_get_transient_id_by_name\n");
  if (yac_couple_config_get_transient_id_by_name(
        couple_config, "dummy_field_2") != 7)
    PUT_ERR("ERROR in yac_couple_config_get_transient_id_by_name\n");
  if (yac_couple_config_get_transient_id_by_name(
        couple_config, "dummy_field_3") != 8)
    PUT_ERR("ERROR in yac_couple_config_get_transient_id_by_name\n");
  if (yac_couple_config_get_transient_id_by_name(
        couple_config, "dummy_field_4") != 9)
    PUT_ERR("ERROR in yac_couple_config_get_transient_id_by_name\n");

  if (!yac_couple_config_transient_config_is_valid(couple_config, 1, 1, 1))
    PUT_ERR("ERROR in yac_couple_config_transient_config_is_valid\n");
  if (!yac_couple_config_transient_config_is_valid(couple_config, 1, 1, 2))
    PUT_ERR("ERROR in yac_couple_config_transient_config_is_valid\n");
  if (!yac_couple_config_transient_config_is_valid(couple_config, 1, 1, 3))
    PUT_ERR("ERROR in yac_couple_config_transient_config_is_valid\n");
  if (!yac_couple_config_transient_config_is_valid(couple_config, 1, 2, 4))
    PUT_ERR("ERROR in yac_couple_config_transient_config_is_valid\n");
  if (!yac_couple_config_transient_config_is_valid(couple_config, 1, 2, 5))
    PUT_ERR("ERROR in yac_couple_config_transient_config_is_valid\n");
  if (!yac_couple_config_transient_config_is_valid(couple_config, 2, 3, 1))
    PUT_ERR("ERROR in yac_couple_config_transient_config_is_valid\n");
  if (!yac_couple_config_transient_config_is_valid(couple_config, 2, 3, 2))
    PUT_ERR("ERROR in yac_couple_config_transient_config_is_valid\n");
  if (!yac_couple_config_transient_config_is_valid(couple_config, 2, 3, 3))
    PUT_ERR("ERROR in yac_couple_config_transient_config_is_valid\n");
  if (!yac_couple_config_transient_config_is_valid(couple_config, 2, 3, 4))
    PUT_ERR("ERROR in yac_couple_config_transient_config_is_valid\n");
  if (!yac_couple_config_transient_config_is_valid(couple_config, 2, 3, 5))
    PUT_ERR("ERROR in yac_couple_config_transient_config_is_valid\n");
  if (yac_couple_config_transient_config_is_valid(couple_config, 1, 3, 1))
    PUT_ERR("ERROR in yac_couple_config_transient_config_is_valid\n");
  if (yac_couple_config_transient_config_is_valid(couple_config, 1, 1, 6))
    PUT_ERR("ERROR in yac_couple_config_transient_config_is_valid\n");
  if (yac_couple_config_transient_config_is_valid(couple_config, 2, 1, 1))
    PUT_ERR("ERROR in yac_couple_config_transient_config_is_valid\n");
  if (yac_couple_config_transient_config_is_valid(couple_config, 1, 5, 1))
    PUT_ERR("ERROR in yac_couple_config_transient_config_is_valid\n");

  if (yac_couple_config_get_timestep_unit(couple_config) != C_SECOND)
    PUT_ERR("ERROR in yac_couple_config_get_timestep_unit\n");

  if (yac_couple_config_get_calendar(couple_config) != PROLEPTIC_GREGORIAN)
    PUT_ERR("ERROR in yac_couple_config_get_calendar");

  if (strcmp(yac_couple_config_get_start_datetime(couple_config),
             "2008-03-09T16:05:07"))
    PUT_ERR("ERROR in yac_couple_config_get_start_datetime\n");

  if (strcmp(yac_couple_config_get_end_datetime(couple_config),
             "2008-03-10T16:05:07"))
    PUT_ERR("ERROR in yac_couple_config_get_end_datetime\n");

  yac_couple_config_set_datetime(
    couple_config, "2009-03-09T16:05:07", "2009-03-10T16:05:07");

  if (strcmp(yac_couple_config_get_start_datetime(couple_config),
             "2009-03-09T16:05:07"))
    PUT_ERR("ERROR in yac_couple_config_get_start_datetime\n");

  if (strcmp(yac_couple_config_get_end_datetime(couple_config),
             "2009-03-10T16:05:07"))
    PUT_ERR("ERROR in yac_couple_config_get_end_datetime\n");

  size_t num_couples = yac_couple_config_get_num_couples(couple_config);
  if (num_couples != 2)
    PUT_ERR("ERROR in yac_couple_config_get_num_couples\n");

  char const * ref_transient_grid_names[1][5][2] =
    {{{"grid1","grid3"},
      {"grid3","grid1"},
      {"grid1","grid3"},
      {"grid3","grid2"},
      {"grid2","grid3"}}};
  int ref_transient_component_ids[1][5][2] = {{{1,2},{2,1},{1,2},{2,1},{1,2}}};
  size_t ref_transient_collection_size[1][5] = {{3,3,4,4,5}};
  char const * ref_transient_name[1][5] =
    {{"sea_surface_temperature",
      "wind_speed",
      "water_flux_into_sea_water",
      "grid_eastward_wind",
      "grid_northward_wind"}};
  int ref_mapping_on_source[1][5] = {{1,1,0,0,1}};
  int ref_coupling_period[1][5] = {{10,20,30,40,50}};
  enum yac_reduction_type ref_coupling_period_operation[1][5] =
    {{INSTANT, ACCUMULATE, TIME_AVERAGE, TIME_MINIMUM, TIME_MAXIMUM}};
  int ref_timestep[1][5][2] = {{{1,10},{2,20},{3,30},{4,40},{5,50}}};
  int ref_lag[1][5][2] = {{{0,4},{1,3},{2,2},{3,1},{4,0}}};
  int ref_enforce_write_weight_file[1][5] = {{1,0,1,0,1}};
  char const * ref_weight_file_name[1][5] =
    {{"weights1.nc",
      "weights2.nc",
      "weights3.nc",
      "weights4.nc",
      "weights5.nc"}};
  struct yac_interp_stack_config ref_interp_stack_config[1][5] =
    {{(struct yac_interp_stack_config){
         .config =
           (union yac_interp_stack_config_entry[6])
             {(union yac_interp_stack_config_entry){.n_nearest_neighbor.type = N_NEAREST_NEIGHBOR,
                                                    .n_nearest_neighbor.config.type = NNN_AVG,
                                                    .n_nearest_neighbor.config.n = 16},
              (union yac_interp_stack_config_entry){.average.type = AVERAGE,
                                                    .average.reduction_type = AVG,
                                                    .average.partial_coverage = 0},
              (union yac_interp_stack_config_entry){.conservative.type = CONSERVATIVE,
                                                    .conservative.order = 1,
                                                    .conservative.enforced_conserv = 0,
                                                    .conservative.partial_coverage = 0,
                                                    .conservative.normalisation = CONSERV_DESTAREA},
              (union yac_interp_stack_config_entry){.hcsbb.type = BERNSTEIN_BEZIER},
              (union yac_interp_stack_config_entry){.user_file.type = USER_FILE,
                                                    .user_file.filename = "weights.nc",
                                                    .user_file.src_grid_name = "grid1",
                                                    .user_file.tgt_grid_name = "grid3",
                                                    .user_file.io_proc_step = -1},
              (union yac_interp_stack_config_entry){.fixed.type = FIXED_VALUE,
                                                    .fixed.value = -1.0}},
         .size = 6},
      (struct yac_interp_stack_config){
         .config =
           (union yac_interp_stack_config_entry[3])
             {(union yac_interp_stack_config_entry){.average.type = AVERAGE,
                                                    .average.reduction_type = AVG,
                                                    .average.partial_coverage = 0},
              (union yac_interp_stack_config_entry){.n_nearest_neighbor.type = N_NEAREST_NEIGHBOR,
                                                    .n_nearest_neighbor.config.type = NNN_DIST,
                                                    .n_nearest_neighbor.config.n = 2},
              (union yac_interp_stack_config_entry){.conservative.type = CONSERVATIVE,
                                                    .conservative.order = 2,
                                                    .conservative.enforced_conserv = 1,
                                                    .conservative.partial_coverage = 1,
                                                    .conservative.normalisation = CONSERV_FRACAREA}},
         .size = 3},
      (struct yac_interp_stack_config){
         .config =
           (union yac_interp_stack_config_entry[5])
             {(union yac_interp_stack_config_entry){.check.type = CHECK,
                                                    .check.constructor_key = "",
                                                    .check.do_search_key = ""},
              (union yac_interp_stack_config_entry){.check.type = CHECK,
                                                    .check.constructor_key = "check_constructor",
                                                    .check.do_search_key = ""},
              (union yac_interp_stack_config_entry){.check.type = CHECK,
                                                    .check.constructor_key = "",
                                                    .check.do_search_key = "check_do_search"},
              (union yac_interp_stack_config_entry){.check.type = CHECK,
                                                    .check.constructor_key = "check_constructor",
                                                    .check.do_search_key = "check_do_search"},
              (union yac_interp_stack_config_entry){.radial_basis_function.type = RADIAL_BASIS_FUNCTION,
                                                    .radial_basis_function.config.type = NNN_RBF,
                                                    .radial_basis_function.config.n = 4,
                                                    .radial_basis_function.config.data.rbf_scale =
                                                      YAC_DEFAULT_RBF_SCALE}},
         .size = 5},
      (struct yac_interp_stack_config){
         .config =
           (union yac_interp_stack_config_entry[2])
             {(union yac_interp_stack_config_entry){.n_nearest_neighbor.type = N_NEAREST_NEIGHBOR,
                                                    .n_nearest_neighbor.config.type = NNN_GAUSS,
                                                    .n_nearest_neighbor.config.n = 8,
                                                    .n_nearest_neighbor.config.data.gauss_scale = 0.2},
              (union yac_interp_stack_config_entry){.spmap.type = SOURCE_TO_TARGET_MAP,
                                                    .spmap.spread_distance = YAC_DEFAULT_SPREAD_DISTANCE,
                                                    .spmap.max_search_distance = YAC_DEFAULT_MAX_SEARCH_DISTANCE,
                                                    .spmap.weight_type = SPMAP_DIST}},
         .size = 2},
      (struct yac_interp_stack_config){
         .config =
           (union yac_interp_stack_config_entry[3])
             {(union yac_interp_stack_config_entry){.creep.type = CREEP,
                                                    .creep.creep_distance = -1.0},
              (union yac_interp_stack_config_entry){.user_callback.type = USER_CALLBACK,
                                                    .user_callback.func_compute_weights_key = "compute_weights"},
              (union yac_interp_stack_config_entry){.fixed.type = FIXED_VALUE,
                                                    .fixed.value = -1.0}},
         .size = 3}}};

  for (size_t couple_idx = 0; couple_idx < 1; ++couple_idx) {

    int couple_comp_ids[2];
    yac_couple_config_get_couple_component_ids(
      couple_config, couple_idx, couple_comp_ids);

    if ((couple_comp_ids[0] != 1) || (couple_comp_ids[1] != 2))
      PUT_ERR("ERROR in yac_couple_config_get_couple_component_ids\n");

    size_t num_transient_couples =
      yac_couple_config_get_num_couple_transients(couple_config, couple_idx);

    if (num_transient_couples != 5)
      PUT_ERR("ERROR in yac_couple_config_get_num_couple_transients\n");

    for (size_t transient_couple_idx = 0;
         transient_couple_idx < num_transient_couples; ++transient_couple_idx) {

      char const * src_grid_name, * tgt_grid_name;
      yac_couple_config_get_transient_grid_names(
        couple_config, couple_idx, transient_couple_idx,
        &src_grid_name, &tgt_grid_name);

      if (strcmp(src_grid_name,
                 ref_transient_grid_names[couple_idx][transient_couple_idx][0]) ||
          strcmp(tgt_grid_name,
                 ref_transient_grid_names[couple_idx][transient_couple_idx][1]))
        PUT_ERR("ERROR in yac_couple_config_get_transient_grid_names\n");

      int src_component_id, tgt_component_id;
      yac_couple_config_get_transient_couple_component_ids(
        couple_config, couple_idx, transient_couple_idx,
        &src_component_id, &tgt_component_id);

      if ((src_component_id !=
           ref_transient_component_ids[couple_idx][transient_couple_idx][0]) ||
          (tgt_component_id !=
           ref_transient_component_ids[couple_idx][transient_couple_idx][1]))
        PUT_ERR("ERROR in yac_couple_config_get_transient_component_ids\n");

      if (yac_couple_config_get_transient_collection_size(
            couple_config, couple_idx, transient_couple_idx) !=
          ref_transient_collection_size[couple_idx][transient_couple_idx])
        PUT_ERR("ERROR in yac_couple_config_get_transient_collection_size\n");

      char const * transient_name =
        yac_couple_config_get_transient_name(
          couple_config, couple_idx, transient_couple_idx);
      if (strcmp(
            transient_name, ref_transient_name[couple_idx][transient_couple_idx]))
        PUT_ERR("ERROR in yac_couple_config_get_transient_name\n");

      if (yac_couple_config_mapping_on_source(
            couple_config, couple_idx, transient_couple_idx) !=
          ref_mapping_on_source[couple_idx][transient_couple_idx])
        PUT_ERR("ERROR in yac_couple_config_mapping_on_source\n");

      if (atoi(
            yac_couple_config_get_coupling_period(
              couple_config, couple_idx, transient_couple_idx)) !=
          ref_coupling_period[couple_idx][transient_couple_idx])
        PUT_ERR("ERROR in yac_couple_config_get_coupling_period\n");

      if (yac_couple_config_get_coupling_period_operation(
            couple_config, couple_idx, transient_couple_idx) !=
          ref_coupling_period_operation[couple_idx][transient_couple_idx])
        PUT_ERR("ERROR in yac_couple_config_get_coupling_period_operation\n");

      char const * src_timestep;
      int src_lag;
      yac_couple_config_get_source_timestep(
        couple_config, couple_idx, transient_couple_idx, &src_timestep, &src_lag);

      if ((atoi(src_timestep) != ref_timestep[couple_idx][transient_couple_idx][0]) ||
          (src_lag != ref_lag[couple_idx][transient_couple_idx][0]))
        PUT_ERR("ERROR in yac_couple_config_get_source_timestep\n");

      char const * tgt_timestep;
      int tgt_lag;
      yac_couple_config_get_target_timestep(
        couple_config, couple_idx, transient_couple_idx, &tgt_timestep, &tgt_lag);

      if ((atoi(tgt_timestep) != ref_timestep[couple_idx][transient_couple_idx][1]) ||
          (tgt_lag != ref_lag[couple_idx][transient_couple_idx][1]))
        PUT_ERR("ERROR in yac_couple_config_get_target_timestep\n");

      if (yac_couple_config_enforce_write_weight_file(
            couple_config, couple_idx, transient_couple_idx) !=
          ref_enforce_write_weight_file[couple_idx][transient_couple_idx])
        PUT_ERR("ERROR in yac_couple_config_enforce_write_weight_file\n");

      char const * weight_file_name =
        yac_couple_config_get_weight_file_name(
          couple_config, couple_idx, transient_couple_idx);
      if (strcmp(
            weight_file_name, ref_weight_file_name[couple_idx][transient_couple_idx]))
        PUT_ERR("ERROR in yac_couple_config_get_weight_file_name\n");

      struct yac_interp_stack_config interp_stack_config =
        yac_couple_config_get_interp_stack(
          couple_config, couple_idx, transient_couple_idx);

      if (interp_stack_config.size !=
          ref_interp_stack_config[couple_idx][transient_couple_idx].size)
        PUT_ERR("ERROR in yac_couple_config_get_interp_stack\n");

      if (yac_interp_stack_config_compare(
             &(ref_interp_stack_config[couple_idx][transient_couple_idx]),
             &interp_stack_config))
        PUT_ERR("ERROR in yac_interp_stack_config_compare");

      for (size_t method_idx = 0; method_idx < interp_stack_config.size;
           ++method_idx) {

        if (yac_interp_stack_config_entry_compare(
              &(ref_interp_stack_config[couple_idx][transient_couple_idx]
                  .config[method_idx]),
              &(interp_stack_config.config[method_idx])))
          PUT_ERR("ERROR in yac_couple_config_get_interp_stack or "
                  "yac_interp_stack_config_entry_compare");
      }
    }
  }

  if (yac_couple_config_get_free_grid_id(couple_config) != 5)
    PUT_ERR("ERROR(yac_couple_config_get_free_grid_id): wrong grid id");
  if (yac_couple_config_get_free_grid_id(couple_config) != 5)
    PUT_ERR("ERROR(yac_couple_config_get_free_grid_id): wrong grid id");
  if (yac_couple_config_get_free_component_id(couple_config) != 4)
    PUT_ERR("ERROR(yac_couple_config_get_free_component_id): wrong component id");
  if (yac_couple_config_get_free_component_id(couple_config) != 4)
    PUT_ERR("ERROR(yac_couple_config_get_free_component_id): wrong component id");
  if (yac_couple_config_get_free_transient_id(couple_config) != 10)
    PUT_ERR("ERROR(yac_couple_config_get_free_transient_id): wrong transiehnt id");
  if (yac_couple_config_get_free_transient_id(couple_config) != 10)
    PUT_ERR("ERROR(yac_couple_config_get_free_transient_id): wrong transiehnt id");

  yac_couple_config_delete(couple_config);
}

static struct yac_couple_config * generate_couple_config_via_UI() {

  struct yac_couple_config * couple_config = yac_couple_config_new();

  yac_couple_config_set_timestep_unit(couple_config, C_SECOND);
  yac_couple_config_set_datetime(
    couple_config, "2008-03-09T16:05:07", "2008-03-10T16:05:07");
  yac_couple_config_set_calendar(couple_config, PROLEPTIC_GREGORIAN);

  yac_couple_config_add_grid(couple_config, "grid1", 1);
  yac_couple_config_add_grid(couple_config, "grid2", 2);
  yac_couple_config_add_grid(couple_config, "grid3", 3);
  yac_couple_config_add_grid(couple_config, "grid4", 4);

  yac_couple_config_add_component(couple_config, "ICON-O", 1);
  yac_couple_config_add_component(couple_config, "ICON-A", 2);
  yac_couple_config_add_component(couple_config, "DUMMY", 5);
  yac_couple_config_add_component(couple_config, "DUMMY_2", 3);

  yac_couple_config_add_transient(
    couple_config, "sea_surface_temperature", 1, 3);
  yac_couple_config_add_transient(
    couple_config, "wind_speed", 2, 3);
  yac_couple_config_add_transient(
    couple_config, "water_flux_into_sea_water", 3, 4);
  yac_couple_config_add_transient(
    couple_config, "grid_eastward_wind", 4, 4);
  yac_couple_config_add_transient(
    couple_config, "grid_northward_wind", 5, 5);
  yac_couple_config_add_transient(
    couple_config, "dummy_field_1", 6, 1);
  yac_couple_config_add_transient(
    couple_config, "dummy_field_2", 7, 1);
  yac_couple_config_add_transient(
    couple_config, "dummy_field_3", 8, 1024);
  yac_couple_config_add_transient(
    couple_config, "dummy_field_4", 9, 1024);

  yac_couple_config_set_transient_collection_size(
    couple_config, "dummy_field_3", 1);
  yac_couple_config_set_transient_collection_size(
    couple_config, "dummy_field_4", 1);

  struct yac_interp_stack_config interp_stack = yac_interp_stack_config_new();
  yac_interp_stack_config_add_nnn(&interp_stack, NNN_AVG, 16, 0.0);
  yac_interp_stack_config_add_average(&interp_stack, AVG, 0);
  yac_interp_stack_config_add_conservative(&interp_stack, 1, 0, 0, CONSERV_DESTAREA);
  yac_interp_stack_config_add_hcsbb(&interp_stack);
  yac_interp_stack_config_add_user_file(
    &interp_stack, "weights.nc", "grid1", "grid3", -1);
  yac_interp_stack_config_add_fixed(&interp_stack, -1.0);

  yac_couple_config_add_transient_couple(
    couple_config, 1, 1, 1, "1", 0, 2, 3, "10", 4,
    1, interp_stack, INSTANT, "10", 1, "weights1.nc");

  yac_interp_stack_config_delete(interp_stack);

  interp_stack = yac_interp_stack_config_new();
  yac_interp_stack_config_add_average(&interp_stack, AVG, 0);
  yac_interp_stack_config_add_nnn(&interp_stack, NNN_DIST, 2, 0.0);
  yac_interp_stack_config_add_conservative(&interp_stack, 2, 1, 1, CONSERV_FRACAREA);

  yac_couple_config_add_transient_couple(
    couple_config, 2, 2, 3, "2", 1, 1, 1, "20", 3,
    1, interp_stack, ACCUMULATE, "20", 0, "weights2.nc");

  yac_interp_stack_config_delete(interp_stack);

  interp_stack = yac_interp_stack_config_new();
  yac_interp_stack_config_add_check(
    &interp_stack, "", "");
  yac_interp_stack_config_add_check(
    &interp_stack, "check_constructor", "");
  yac_interp_stack_config_add_check(
    &interp_stack, "", "check_do_search");
  yac_interp_stack_config_add_check(
    &interp_stack, "check_constructor", "check_do_search");
  yac_interp_stack_config_add_nnn(
    &interp_stack, NNN_RBF, 4, YAC_DEFAULT_RBF_SCALE);

  yac_couple_config_add_transient_couple(
    couple_config, 3, 1, 1, "3", 2, 2, 3, "30", 2,
    0, interp_stack, TIME_AVERAGE, "30", 1, "weights3.nc");

  yac_interp_stack_config_delete(interp_stack);

  interp_stack = yac_interp_stack_config_new();
  yac_interp_stack_config_add_nnn(&interp_stack, NNN_GAUSS, 8, 0.2);
  yac_interp_stack_config_add_spmap(&interp_stack, 0.0, 0.0, SPMAP_DIST);

  yac_couple_config_add_transient_couple(
    couple_config, 4, 2, 3, "4", 3, 1, 2, "40", 1,
    0, interp_stack, TIME_MINIMUM, "40", 0, "weights4.nc");

  yac_interp_stack_config_delete(interp_stack);

  interp_stack = yac_interp_stack_config_new();
  yac_interp_stack_config_add_creep(&interp_stack, -1.0);
  yac_interp_stack_config_add_user_callback(&interp_stack, "compute_weights");
  yac_interp_stack_config_add_fixed(&interp_stack, -1.0);

  yac_couple_config_add_transient_couple(
    couple_config, 5, 1, 2, "5", 4, 2, 3, "50", 0,
    1, interp_stack, TIME_MAXIMUM, "50", 1, "weights5.nc");

  yac_couple_config_add_transient_couple(
    couple_config, 5,
    1, 2, "5", 4, // same source
    3, 4, "50", 0, // different target
    1, interp_stack, TIME_MAXIMUM, "50", 1, "weights6.nc");

  yac_interp_stack_config_delete(interp_stack);

  return couple_config;
}
